# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx

from .neural_network_repurposer import NeuralNetworkRepurposer
from .model_handler.model_handler import ModelHandler
from .model_handler import layer_factory
from .constants import neural_network_repurposer_keys as keys


class NeuralNetworkRandomFreezeRepurposer(NeuralNetworkRepurposer):
    """
    Class that creates a target neural network from a source neural network through Transfer Learning.
    It transfers layers from source model and fine-tunes the network to learn from target data set.

    Steps involved:

    - Layers and weights of source model are transferred
    - N layers are removed from the output of the model
    - A fully connected layer with nodes equal to number of classes in target data set is added to the model output
    - A softmax layer is added on top of the new fully connected layer
    - Some layers are frozen and some randomly initialized and then model is fine-tuned

    :param source_model: Source neural network to do transfer leaning from. Can be None in a predict-only case
    :type source_model: :class:`mxnet.mod.Module`
    :param int target_class_count: Number of classes to train the target neural network for. Can be None in a
                                   predict-only case
    :param list[str] fixed_layers: List of layers to keep weights frozen for
    :param list[str] random_layers: List of layers to randomly reinitialise
    :param int num_layers_to_drop: Number of layers to remove from model output
    :param context_function: MXNet context function that provides device type context
    :type context_function: function
    :param int num_devices: Number of devices to use to train target neural network
    :param int batch_size: Size of data batches to be used for training the target neural network
    :param int num_epochs: Number of epochs to be used for training the target neural network
    :param str optimizer: Optimizer required by MXNet to train target neural network. Default: 'sgd'
    :param optimizer_params: Optimizer params required by MXNet to train target neural network.
           Default: {'learning_rate': 1e-3}
    :type optimizer_params: dict(str, float)
    """
    def __init__(self, source_model: mx.mod.Module, target_class_count, fixed_layers, random_layers,
                 num_layers_to_drop=2, context_function=mx.context.cpu, num_devices=1, batch_size=64, num_epochs=5,
                 optimizer='sgd', optimizer_params=None):
        super().__init__(source_model, context_function, num_devices, batch_size, num_epochs,
                         optimizer, optimizer_params)

        self.target_class_count = target_class_count
        self.fixed_layers = fixed_layers
        self.random_layers = random_layers
        self.num_layers_to_drop = num_layers_to_drop

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor

        :rtype: dict
        """
        param_dict = super().get_params()
        param_dict[keys.TARGET_CLASS_COUNT] = self.target_class_count
        param_dict[keys.FIXED_LAYERS] = self.fixed_layers
        param_dict[keys.RANDOM_LAYERS] = self.random_layers
        param_dict[keys.NUM_LAYERS_TO_DROP] = self.num_layers_to_drop
        return param_dict

    def _create_target_module(self, train_iterator: mx.io.DataIter):
        """
        Function to create target module by transferring layers and weights of source model, removing layers from output
        and freezing and randomising selected layers

        Overrides abstract method in base NeuralNetworkRepurposer class. Called by repurpose method in base class to
        create target MXNet module that needs to be trained with target data set.

        :param train_iterator: Training data iterator. Used by this function to bind target module
        :type train_iterator: :class: `mxnet.io.DataIter`
        :return: Target MXNet module created by transferring layers from source model
        :rtype: :class:`mxnet.mod.Module`
        """
        # Create model handler to manipulate the source model
        model_handler = ModelHandler(self.source_model, self.context_function, self.num_devices)

        # Check if random and frozen layers are present in source model
        assert type(self.fixed_layers) == list, 'fixed_layers should be a list'
        assert type(self.random_layers) == list, 'random_layers should be a list'
        for layer_name in self.fixed_layers + self.random_layers:
            if layer_name not in model_handler.layer_names:
                raise ValueError('layer name: {} not found in source model'.format(layer_name))

        # Get lists of parameters to fix and randomise
        fixed_params = model_handler.get_layer_parameters(self.fixed_layers)
        random_params = model_handler.get_layer_parameters(self.random_layers)

        # Drop layers
        model_handler.drop_layer_top(self.num_layers_to_drop)

        # Add fully connected layer and softmax output
        fc = layer_factory.FullyConnected(name='new_fully_connected_layer', num_hidden=self.target_class_count)
        # Softmax layer name should be set to the name provided by the iterator
        softmax_name = train_iterator.provide_label[0][0].replace('_label', '')
        softmax = layer_factory.SoftmaxOutput(name=softmax_name)
        model_handler.add_layer_top([fc, softmax])

        # Create and return target MXNet module using the fixed/random parameters
        return model_handler.get_module(train_iterator, fixed_layer_parameters=fixed_params,
                                        random_layer_parameters=random_params)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import json
import logging
import gc
import mxnet as mx
import numpy as np
from collections import OrderedDict

from . import exceptions, consts, layer_factory


class ModelHandler(object):
    """
    Class for model manipulation and feature extraction.

    :param module: MXNet module to be manipulated.
    :type module: :class:`mx.module.Module`
    :param function context_function: MXNet context function.
    :param int num_devices: Number of devices to run process on.
    :param str data_name: Name of input layer of model.
    """
    def __init__(self, module, context_function=mx.context.cpu, num_devices=1, data_name=consts.DATA):
        self.symbol = module.symbol
        if module.binded and module.params_initialized:
            self.arg_params, self.aux_params = module.get_params()
        else:
            self.arg_params = module._arg_params if module._arg_params is not None else {}
            self.aux_params = module._aux_params if module._aux_params is not None else {}

        self.layer_type_dict = self._get_layer_type_dict()
        self.devices = self._get_devices(context_function, num_devices)
        self.data_name = data_name

    def drop_layer_top(self, num_layers_to_drop=1):
        """
        Remove layers from output of model.

        :param int n: Number of layers to remove from model output.
        """
        self._assert_drop_layer_valid(num_layers_to_drop)
        # Get index of new last layer
        last_layer_idx = -(num_layers_to_drop+1)
        # symbol ending at new last layer
        sym = self.symbol.get_internals()[self.layer_names[last_layer_idx] + '_output']
        logging.info('{} deleted from model top'.format(', '.join(self.layer_names[last_layer_idx + 1:])))
        self.update_sym(sym)

    def drop_layer_bottom(self, num_layers_to_drop=1):
        """
        Remove layers from input of model.
        This method requires the entire symbol to be recreated internally.

        :param int n: Number of layers to remove from model input.
        """
        self._assert_drop_layer_valid(num_layers_to_drop)
        symbol_nodes = self._get_symbol_dict()[consts.NODES]
        # Get index of nth layer
        layer_count = 0
        n_layer_idx = 0
        for id, node in enumerate(symbol_nodes):
            if node[consts.OPERATION] != consts.NO_OP:
                layer_count += 1
            if layer_count == num_layers_to_drop:
                n_layer_idx = id
                break
        # Raise error if nth layer is not found
        if n_layer_idx == 0:
            raise exceptions.ModelError('{}th layer could not be found'.format(num_layers_to_drop))
        # Remove nodes in list that relate to first n layers (keep first node because this is data layer)
        symbol_nodes = symbol_nodes[0:1] + symbol_nodes[n_layer_idx+1:]
        id2name = {count: node[consts.LAYER_NAME] for count, node in enumerate(symbol_nodes)}
        # Recreate model from symbol nodes
        sym = self._model_from_nodes(None, symbol_nodes, elements_offset=-n_layer_idx, id2name=id2name,
                                     data_name=self.data_name)
        logging.info('{} deleted from model bottom'.format(', '.join(self.layer_names[:num_layers_to_drop])))
        self.update_sym(sym)

    def add_layer_top(self, layer_factory_list):
        """
        Add layer to output of model.

        :param layer_factory: List of LayerFactory objects to be added to model output.
        :type layer_factory: list(:class:`LayerFactory`)
        """
        if '_label' in self.symbol.get_internals().list_outputs()[consts.LABEL_IDX]:
            raise exceptions.ModelError('Cannot add layer above output layer')
        sym = self.symbol
        added_layer_names = []
        for lf in layer_factory_list:
            self._validate_layer_name(lf.attributes[consts.NAME])
            sym = lf.create_layer(sym)
            added_layer_names.append(lf.attributes[consts.NAME])
        self.update_sym(sym)
        logging.info('Added {} to model top'.format(', '.join(added_layer_names)))

    def add_layer_bottom(self, layer_factory_list):
        """
        Add layer to input of model.
        This method requires the entire symbol to be recreated internally.

        :param layer_factory: List of LayerFactory objects to be added to model input.
        :type layer_factory: list(:class:`LayerFactory`)
        """
        # Create data layer
        symbol_nodes = self._get_symbol_dict()[consts.NODES]
        data_layer = layer_factory.Data(name=self.data_name)
        sym = data_layer.create_layer(None)
        prev_symbols = {self.data_name: sym}
        id2name = {0: self.data_name}
        added_layer_names = []
        # Add new layers
        for id, lf in enumerate(layer_factory_list):
            self._validate_layer_name(lf.attributes[consts.NAME])
            if lf.output is True:
                raise exceptions.ModelError('Cannot add output layer to model bottom ({})'
                                            .format(lf.attributes[consts.NAME]))
            sym = lf.create_layer(sym)
            prev_symbols[lf.attributes[consts.NAME]] = sym
            # Update id2name with newly added layer
            id2name[id + 1] = lf.attributes[consts.NAME]
            added_layer_names.append(lf.attributes[consts.NAME])
        # Add remaining symbol names to id2name (beginning slice at 1 to skip data node which is already in id2name)
        for count, node in enumerate(symbol_nodes[1:]):
            id2name[count + len(layer_factory_list) + 1] = node[consts.LAYER_NAME]
        # Recreate model
        sym = self._model_from_nodes(sym, symbol_nodes, elements_offset=len(layer_factory_list),
                                     prev_symbols=prev_symbols, id2name=id2name, data_name=self.data_name)
        self.update_sym(sym)
        logging.info('Added {} to model bottom'.format(', '.join(added_layer_names)))

    def get_module(self, iterator, fixed_layer_parameters=None, random_layer_parameters=None):
        """
        Return MXNet Module using the model symbol and parameters.

        :param iterator: MXNet iterator to be used with model.
        :type iterator: :class:`mxnet.io.DataIter`
        :param list(str) fixed_layer_parameters: List of layer parameters to keep fixed.
        :param list(str) random_layer_parameters: List of layer parameters to randomise.
        :return: MXNet module
        :rtype: :class:`mx.module.Module`
        """
        if fixed_layer_parameters is not None:
            fixed_layer_parameters = self._prune_parameters(fixed_layer_parameters)
        if random_layer_parameters is None:
            arg_params, aux_params = self.arg_params.copy(), self.aux_params.copy()
        else:
            arg_params, aux_params = self._remove_random_parameters(random_layer_parameters)
        mod = mx.mod.Module(symbol=self.symbol, context=self.devices, fixed_param_names=fixed_layer_parameters,
                            label_names=(self.layer_names[-1] + "_label",), data_names=(self.data_name,))
        mod.bind(data_shapes=iterator.provide_data, label_shapes=iterator.provide_label)
        mod.init_params(mx.init.Xavier(rnd_type='gaussian', factor_type='in', magnitude=2))
        try:
            mod.set_params(arg_params, aux_params, allow_missing=True, force_init=True)
        except mx.MXNetError as e:
            exceptions._handle_mxnet_error(e)
        return mod

    def get_layer_type(self, layer_name):
        """
        Return type of named layer.

        :param str name: Name of layer being inspected.
        :return: Layer type
        :rtype: str
        """
        try:
            return self.layer_type_dict[layer_name]
        except KeyError:
            raise ValueError('{} layer does not exist in model'.format(layer_name))

    def get_layer_names_matching_type(self, layer_type):
        """
        Return names of layers of specified type.

        :param LayerType layer_type: Return list of layers of this type. Should be a LayerType enum.
        :return: Names of layers with specified type
        :rtype: list(str)
        """
        return [layer_name for layer_name, l_type in self.layer_type_dict.items() if l_type == layer_type.value]

    def get_layer_output(self, data_iterator, layer_names):
        """
        Function to extract features from data iterator with model.
        Returns a dictionary of layer_name -> numpy array of features extracted (flattened).

        :param data_iterator: Iterator containing input data.
        :type data_iterator: :class:`mxnet.io.DataIter`
        :param list(str) layer_names: List of names of layers to extract features from.
        :return: Ordered Dictionary of features ({layer_name: features}), list of labels
                 Layer names in the ordered dictionary follow the same order as input list of layer_names
        :rtype: OrderedDict[str, :class:`numpy.array`], list(int)
        """
        if not isinstance(data_iterator, mx.io.DataIter):
            raise TypeError('Iterator should be an MXNet DataIter object'.format(type(data_iterator)))
        if type(layer_names) is not list:
            raise TypeError('Layers must be passed as list.')
        logging.info('Extracting features from layers: ' + ' '.join(layer_names))

        features = OrderedDict()
        labels = np.array([])
        data_shape = data_iterator.provide_data[0][1]
        batch_no = 0
        id2layer = dict(enumerate(layer_names))
        data_iterator.reset()

        intermediate_layer_symbols = [self.symbol.get_internals()[l + '_output'] for l in layer_names]
        intermediate_symbol = mx.sym.Group(intermediate_layer_symbols)

        module = mx.mod.Module(symbol=intermediate_symbol, context=self.devices, label_names=None,
                               data_names=(self.data_name,))
        module.bind(for_training=False, data_shapes=[(self.data_name, data_shape)], label_shapes=None)
        module.set_params(self.arg_params, self.aux_params)

        # MXNet Module has issues releasing memory from GPU (https://github.com/apache/incubator-mxnet/issues/5983)so we
        # force full garbage collection before feature extraction and force a shallow garbage collection after every
        # loop to free up any memory that has been allocated during the loop and should have been released
        gc.collect(2)  # Full garbage collection before begininning to extract features

        while True:
            try:
                batch = data_iterator.next()
            except StopIteration:
                break
            batch_no += 1
            labels = np.append(labels, batch.label[0].asnumpy())
            module.forward(batch)
            forward_output = module.get_outputs()

            for lid in id2layer:
                feature = forward_output[lid].asnumpy().flatten().reshape(data_shape[0], -1)
                # Remove extra values from padding if they are present
                feature = feature[:feature.shape[0]-batch.pad]
                if id2layer[lid] in features:
                    features[id2layer[lid]] = np.vstack((features[id2layer[lid]], feature))
                else:
                    features[id2layer[lid]] = feature
            # Remove extra values from padding if they are present
            labels = labels[:labels.shape[0]-batch.pad]

            logging.info('Processed batch {0}'.format(batch_no))
            del forward_output
            del feature
            del batch
            gc.collect(0)  # Shallow garbage collect for objects created during loop

        return features, labels.astype(int)

    def get_layer_parameters(self, layer_names):
        """
        Get list of layer parameters associated with the the layer names given.

        :param list(str) layer_names: List of layer names.
        :return: List of layer parameters
        :rtype: list(str)
        """
        if type(layer_names) is not list:
            raise TypeError('layer_names must be passed as list.')
        params = []
        for layer_name in layer_names:
            for arg_dict in [self.arg_params, self.aux_params]:
                new_params = [param for param in list(arg_dict.keys()) if layer_name in param]
                params = params + new_params
        return params

    def visualize_net(self):
        """
        Display computational graph of model.
        """
        return mx.viz.plot_network(self.symbol, node_attrs={'fixedsize': 'false'})

    def save_symbol(self, model_name):
        """
        Serialise model symbol graph.

        :param str model_name: Prefix to file name (model_name-symbol.json).
        """
        self.symbol.save(model_name + '-symbol.json')

    def _validate_layer_name(self, layer_name):
        """
        Validate name of layer.

        :param str layer_name: Name to be validated.
        """
        # Input name is not included in layer_names so the check for conflict is done with symbol inputs and layer_names
        if layer_name in self.symbol.list_inputs() or layer_name in self.layer_names:
            raise ValueError("Layer name '{}' conflicts with name already in model.".format(layer_name))
        # MXNet uses these suffixes for specific things so we are avoiding using them in layer names
        for suffix in ['output', 'label', 'weight', 'bias', 'moving_mean', 'moving_var', 'gamma', 'beta']:
            if '_' + suffix in layer_name:
                raise ValueError("Layer name cannot contain '{}'".format(suffix))

    @staticmethod
    def _model_from_nodes(symbol, symbol_nodes, elements_offset=0, prev_symbols=None, id2name=None,
                          data_name=consts.DATA):
        """
        Recreate model from dictionary above symbol.

        :param :class:`LayerFactory` symbol: Symbol on which to create model.
        :param list symbol_nodes: List of dictionaries which desribe model layers.
        :param int elements_offset: Number of layers added to model since symbol dictionary was last updated.
        :param dict(str, :class:`Symbol`) prev_symbols: Dictionary of all MXNet symbols corresponding to previously
                                                        added layers -- layer_name: MXNet symbol.
        :param dict(int, str) id2name: Names of symbols in model -- {position of symbol in model: symbol name}.
        :rtype: :class:`Symbol`
        """
        prev_symbols = {} if prev_symbols is None else prev_symbols
        for count, node in enumerate(symbol_nodes):
            node[consts.ELEMENTS_OFFSET] = elements_offset
            node[consts.PREV_SYMBOLS] = prev_symbols
            node[consts.ID2NAME] = id2name
            node[consts.DATA] = data_name
            new_layer, input_symbol = layer_factory.LayerFactory._from_dict(node)
            # If node does not represent a layer operation then new_layer will be None
            if new_layer:
                symbol = new_layer.create_layer(input_symbol)
                prev_symbols[new_layer.attributes[consts.NAME]] = symbol
        return symbol

    def _prune_parameters(self, parameter_names):
        """
        Remove parameter names from list which are not in model parameter dicts.  Logs warning for removed names.

        :param list(str) parameter_names: List of parameter names.
        :return: Parameter names that are all in the model parameter dicts
        :rtype: list(str)
        """
        all_keys = list(self.arg_params.keys()) + list(self.aux_params.keys())
        names_not_found = [parameter_name for parameter_name in parameter_names if parameter_name not in all_keys]
        if len(names_not_found) > 0:
            logging.warning('Could not find layer parameters: {}'.format(', '.join(names_not_found)))
        return [parameter_name for parameter_name in parameter_names if parameter_name in all_keys]

    def _remove_random_parameters(self, random_parameter_names):
        """
        Return model parameter dicts with random parameters removed.

        :param list(str) random_parameter_names: Names of random parameters to be removed from model parameter dicts.
        :return: arg_params, aux_params
        :rtype: dict, dict
        """
        random_parameter_names = self._prune_parameters(random_parameter_names)
        arg_params = self.arg_params.copy()
        aux_params = self.aux_params.copy()
        for parameter_name in random_parameter_names:
            arg_params.pop(parameter_name, None)
            aux_params.pop(parameter_name, None)
        return arg_params, aux_params

    @staticmethod
    def _get_devices(context_function=mx.context.cpu, num_devices=1):
        """
        Return devices list.

        :param function context_function: MXNet context function which returns a context.
        :param int num_devices: Number of devices to use for processing (e.g number of cpus or gpus to use).
        :return: List of devices
        :rtype: list
        """
        return [context_function(i) for i in range(num_devices)]

    def _assert_drop_layer_valid(self, n):
        """
        Raise exception if the number of layers being dropped is more than or equal to the number of layers in the
        model.
        """
        if len(self.layer_names) < n + 1:
            raise exceptions.ModelError('Cannot drop {} layers. Model only has {} layers'.format(n,
                                                                                                 len(self.layer_names)))

    def _get_layer_type_dict(self):
        """
        Return dictionary of layer types.

        :return: Dictionary of layer names to layer types
        :rtype: dict
        """
        symbol_dict = self._get_symbol_dict()
        # Each layer in the model has one output (along with a bias or weight etc.) so we filter for all the list
        # entries containing '_output' and then we remove this substring from the list entries to leave just the layer
        # names
        layer_names = [l.replace('_output', '') for l in self.symbol.get_internals().list_outputs() if '_output' in l]
        symbol_layers = [node for node in symbol_dict[consts.NODES] if node[consts.LAYER_NAME] in layer_names]
        layer_type_dict = OrderedDict()
        for layer in symbol_layers:
            layer_type_dict[layer[consts.LAYER_NAME]] = layer[consts.OPERATION]
        return layer_type_dict

    def _get_symbol_dict(self):
        """
        Get symbol dictionary.

        :return: Symbol dictionary
        :rtype: dict
        """
        return json.loads(self.symbol.tojson())

    @property
    def layer_names(self):
        """
        Get list of names of model layers.

        :return: List of layer names
        :rtype: list[str]
        """
        return list(self.layer_type_dict.keys())

    @staticmethod
    def _clean_params(symbol, parameter_dict):
        """
        Return a copy of parameter_dict with parameters for layers that are not in the symbol removed.

        :param symbol: Symbol to give point of reference for removing parameters.
        :type symbol: :class:`mx.symbol.Symbol`
        :param dict parameter_dict: Dictionary of model parameters.
        :return: Parameter dictionary with all parameters referring to layer(s) in the symbol
        :rtype: dict
        """
        parameter_dict = parameter_dict.copy()
        keys_to_delete_arg = set(parameter_dict.keys()) - set(symbol.get_internals().list_outputs())
        for key in keys_to_delete_arg:
            del parameter_dict[key]
        return parameter_dict

    def update_sym(self, new_symbol):
        """
        Update symbol attribute, layer names, and layer types dict and clean parameters.

        :param new_symbol: Symbol with which to update ModelHandler.
        :type new_symbol: :class:`mx.symbol.Symbol`
        """
        self.symbol = new_symbol
        self.layer_type_dict = self._get_layer_type_dict()
        self.arg_params = self._clean_params(self.symbol, self.arg_params)
        self.aux_params = self._clean_params(self.symbol, self.aux_params)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx
from sklearn.linear_model import LogisticRegression

from .meta_model_repurposer import MetaModelRepurposer
from .constants import repurposer_keys
from .constants import meta_model_repurposer_keys as keys
from . import utils


class LrRepurposer(MetaModelRepurposer):
    """
    Perform Transfer Learning through a Logistic Regression meta-model which repurposes the source neural network.

    :param source_model: Source neural network to do Transfer Learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param feature_layer_names: Name of layer(s) in source_model from which features should be transferred.
    :type feature_layer_names: list[str]
    :param context_function: MXnet context function that provides device type context.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use to extract features from source_model.
    :param float tol: Tolerance for stopping criteria.
    :param float c: Inverse of regularization strength; must be a positive float.
    :param int n_jobs: Number of CPU cores used during the cross-validation loop. If given a value of -1, all cores
        are used.
    """
    def __init__(self, source_model: mx.mod.Module, feature_layer_names, context_function=mx.context.cpu, num_devices=1,
                 tol=1e-4, c=1.0, n_jobs=-1):
        # Call base class constructor with parameters required for meta-models
        super(LrRepurposer, self).__init__(source_model, feature_layer_names, context_function, num_devices)

        # Initialize LR specific parameters
        self.tol = tol
        self.c = c
        self.n_jobs = n_jobs

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor.

        :rtype: dict
        """
        param_dict = super(LrRepurposer, self).get_params()
        param_dict[keys.TOL] = self.tol
        param_dict[keys.C] = self.c
        param_dict[keys.N_JOBS] = self.n_jobs
        return param_dict

    def _train_model_from_features(self, features, labels, feature_indices_per_layer=None):
        """
        Train a Logistic Regression model using features extracted from source neural network.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :param labels: Labels to use for training.
        :type labels: :class:`numpy.ndarray`
        :param feature_indices_per_layer: Mapping of feature_layer_names to indices in features array.
        i.e. {layer_name, feature_indices}. Note that this param is currently not consumed by lr_repurposer.
        :type feature_indices_per_layer: OrderedDict[str, :class:`numpy.ndarray`]
        :return: LR model trained with given features and labels using sci-kit learn library.
        :rtype: :class: `sklearn.linear_model.LogisticRegression`
        """
        lr_model = LogisticRegression(penalty='l2',
                                      tol=self.tol,
                                      C=self.c,
                                      fit_intercept=False,
                                      class_weight='balanced',
                                      random_state=1,
                                      solver='sag',
                                      multi_class='multinomial',
                                      n_jobs=self.n_jobs)
        lr_model.fit(features, labels)
        return lr_model

    def _predict_probability_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use sklearn's LR predict_proba method and return predicted probabilities.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted probabilities.
        :rtype: :class:`numpy.ndarray`
        """
        return LogisticRegression.predict_proba(self.target_model, features)

    def _predict_label_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use sklearn's LR predict method and return predicted labels.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted labels.
        :rtype: :class:`numpy.ndarray`
        """
        return LogisticRegression.predict(self.target_model, features)

    def serialize(self, file_prefix):
        """
        Saves repurposer (excluding source model) to file_prefix.json.

        :param str file_prefix: Prefix to save file with.
        """
        output_dict = {}
        output_dict[repurposer_keys.PARAMS] = self.get_params()
        output_dict[repurposer_keys.TARGET_MODEL] = utils.sklearn_model_to_dict(self.target_model)
        output_dict.update(self._get_attributes())

        utils.save_json(file_prefix, output_dict)

    def deserialize(self, input_dict):
        """
        Uses dictionary to set attributes of repurposer.

        :param dict input_dict: Dictionary containing values for attributes to be set to.
        """
        self._set_attributes(input_dict)  # Set attributes of the repurposer from input_dict
        self.target_model = utils.sklearn_model_from_dict(LogisticRegression, input_dict[repurposer_keys.TARGET_MODEL])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
from abc import ABCMeta, abstractmethod
import os
import warnings
import mxnet as mx

from . import utils
from .constants import serialization_keys, constants, serialization_constants
from .constants import repurposer_keys as keys
from .model_handler import ModelHandler


class Repurposer:
    """
    Base Class for repurposers that train models using Transfer Learning (source_model -> target_model).

    :param source_model: Source neural network to do transfer learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param context_function: MXNet context function that provides device type context.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use with context_function.
    """
    __metaclass__ = ABCMeta

    def __init__(self, source_model: mx.mod.Module, context_function=mx.context.cpu, num_devices=1):
        self.source_model = source_model
        self.context_function = context_function
        self.num_devices = num_devices
        self.target_model = None
        self._save_source_model_default = None
        self.provide_data = None
        self.provide_label = None

    @abstractmethod
    def repurpose(self, train_iterator: mx.io.DataIter):
        """
        Abstract method.
        """
        pass

    @abstractmethod
    def predict_probability(self, test_iterator: mx.io.DataIter):
        """
        Abstract method.
        """
        pass

    @abstractmethod
    def predict_label(self, test_iterator: mx.io.DataIter):
        """
        Abstract method.
        """
        pass

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor's argument list.

        :rtype: dict
        """
        param_dict = {}
        param_dict[keys.CONTEXT_FN] = utils.serialize_ctx_fn(self.context_function)
        param_dict[keys.NUM_DEVICES] = self.num_devices
        return param_dict

    def _get_attributes(self):
        """
        Get parameters of repurposer not in constructor's argument list.

        :rtype: dict
        """
        attr_dict = {}
        attr_dict[keys.PROVIDE_DATA] = self.provide_data
        attr_dict[keys.PROVIDE_LABEL] = self.provide_label
        attr_dict[serialization_keys.VERSION] = constants.VERSION
        attr_dict[keys.REPURPOSER_CLASS] = self.__class__.__name__
        attr_dict[serialization_keys.LAST_LAYER_NAME_SOURCE] = ModelHandler(self.source_model).layer_names[-1]
        return attr_dict

    def _set_attributes(self, input_dict):
        """
        Set attributes of class from input_dict.
        These attributes are the same as those returned by get_attributes method.

        :param input_dict: Dictionary containing attribute values.
        :return: None
        """

        # Raise warning if there is a version mismatch with the Xfer version saved in input_dict
        if not constants.VERSION == input_dict[serialization_keys.VERSION]:
            warnings.warn("Running Xfer version: {} which is different from version saved in dictionary: {}. "
                          "Ensure that you are using the correct version of Xfer"
                          .format(constants.VERSION, input_dict[serialization_keys.VERSION]))
        self.provide_data = input_dict[keys.PROVIDE_DATA]
        self.provide_label = input_dict[keys.PROVIDE_LABEL]

    @abstractmethod
    def serialize(self, file_prefix):
        """
        Abstract method.
        """
        pass

    @abstractmethod
    def deserialize(self, input_dict):
        """
        Abstract method.
        """
        pass

    def save_repurposer(self, model_name, model_directory='', save_source_model=None):
        """
        Serialize the repurposed model (source_model, target_model and supporting info) and save it to given file_path.

        :param str model_name: Name to save repurposer to.
        :param str model_directory: File directory to save repurposer in.
        :param boolean save_source_model: Flag to choose whether to save repurposer source model.
                                          Will use default if set to None. (MetaModelRepurposer default: True,
                                          NeuralNetworkRepurposer: False)
        """
        # Assert repurpose() has been called successfully
        if self.target_model is None:
            raise ValueError('Cannot save repurposer before source model has been repurposed')

        file_prefix = os.path.join(model_directory, model_name)

        self._save_source_model(file_prefix, save_source_model)

        # Serialize remainder of repurposer
        self.serialize(file_prefix)

    def _save_source_model(self, file_prefix, save_source_model):
        # Set save_source_model flag to class default if None
        save_source_model = self._save_source_model_default if save_source_model is None else save_source_model
        if save_source_model:
            # save source_model as file-path_source-symbol.json and file-path_source-0000.params
            utils.save_mxnet_model(self.source_model, file_prefix + serialization_constants.SOURCE_SUFFIX, 0,
                                   self.provide_data, self.provide_label)

    def _validate_before_repurpose(self):
        if not isinstance(self.source_model, mx.mod.Module):
            error = ("Cannot repurpose because source_model is not an `mxnet.mod.Module` object."
                     " Instead got type: {}".format(type(self.source_model)))
            raise TypeError(error)

    def _validate_before_predict(self):
        if self.target_model is None:
            raise ValueError("Cannot predict because target_model is not initialized")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet.ndarray as nd
import numpy as np
import mxnet as mx

WRITE = "write"
NULL = "null"
MEAN = "mean"
RHO = "rho"
PARAMS = '_params'
JSON = '.json'
CONTEXT = 'ctx'
SHAPE = 'shape'
NPZ = '.npz'


def log_gaussian(x, mean, sigma):
    return -0.5 * np.log(2.0 * np.pi) - nd.log(sigma) - (x - mean) ** 2 / (2 * sigma ** 2)


def sample_epsilons(param_shapes, ctx):
    return [nd.random_normal(shape=shape, loc=0., scale=1.0, ctx=ctx) for shape in param_shapes]


def transform_gaussian_samples(means, sigmas, epsilons):
    return [means[j] + sigmas[j] * epsilons[j] for j in range(len(epsilons))]


def softplus(x):
    return nd.log(1. + nd.exp(x))


def softplus_inv(x):
    return nd.log(nd.exp(x) - 1.)


def softplus_inv_numpy(x):
    return np.log(np.exp(x) - 1.)


def transform_rhos(rhos):
    return [softplus(rho) for rho in rhos]


def transform_rhos_inverse(sigmas):
    return [softplus_inv(sigma) for sigma in sigmas]


def replace_params_net(layer_params, net, ctx):
    for l_param, param in zip(layer_params, net.collect_params().values()):
        ctx_list = param._ctx_map[ctx.device_typeid & 1]
        if ctx.device_id >= len(ctx_list) or ctx_list[ctx.device_id] is None:
            raise ValueError("Invalid context.")
        dev_id = ctx_list[ctx.device_id]
        param._data[dev_id] = l_param


def serialise_ctx(ctx):
    """
    Convert context to serialisable list
    """
    ctx_list = []
    if type(ctx) is not list:
        ctx = [ctx]
    for c in ctx:
        ctx_list.append((c.device_id, c.device_typeid))
    return ctx_list


def deserialise_ctx(ctx_list):
    """
    Get context from serialised context
    """
    ctx = []
    for ctx_tuple in ctx_list:
        temp_ctx = mx.cpu()
        temp_ctx.device_id = ctx_tuple[0]
        temp_ctx.device_typeid = ctx_tuple[1]
        ctx.append(temp_ctx)
    if len(ctx) == 1:
        ctx = ctx[0]
    return ctx


def deserialise_shape(shape):
    """
    Get shape from serialised shape
    """
    deserialised_shape = []
    for sh in shape:
        deserialised_shape.append(tuple(sh))
    return deserialised_shape



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
from abc import abstractmethod
from collections import OrderedDict
import mxnet as mx
import numpy as np

from .repurposer import Repurposer
from .model_handler import ModelHandler
from .constants import meta_model_repurposer_keys as keys


class MetaModelRepurposer(Repurposer):
    """
    Base class for repurposers that extract features from layers in source neural network (Transfer) and train a
    meta-model using the extracted features (Learn).

    :param source_model: Source neural network to do transfer learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param feature_layer_names: Name of layer(s) in source_model from which features should be transferred.
    :type feature_layer_names: list[str]
    :param context_function: MXNet context function that provides device type context.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use to extract features from source_model.
    """
    def __init__(self, source_model: mx.mod.Module, feature_layer_names, context_function=mx.context.cpu,
                 num_devices=1):
        super(MetaModelRepurposer, self).__init__(source_model, context_function, num_devices)

        # Create model handler for source model. Required to extract features(layer output) during repurpose and predict
        self.source_model_handler = ModelHandler(self.source_model, self.context_function, self.num_devices)

        self.feature_layer_names = feature_layer_names
        self.provide_data = None
        self.provide_label = None
        self._save_source_model_default = True

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor's argument list.

        :rtype: dict
        """
        param_dict = super(MetaModelRepurposer, self).get_params()
        param_dict[keys.FEATURE_LAYERS] = self.feature_layer_names
        return param_dict

    @property
    def feature_layer_names(self):
        """
        Names of the layers to extract features from.
        """
        return self._feature_layer_names

    @feature_layer_names.setter
    def feature_layer_names(self, value):
        if not isinstance(value, list):
            raise TypeError("feature_layer_names must be a list. Instead got type: {}".format(type(value)))

        if len(value) == 0:
            raise ValueError(
                "feature_layer_names cannot be empty. At least one layer name is required to extract features")

        # Validate if feature_layer_names passed are found in source_model
        for layer_name in value:
            if layer_name not in self.source_model_handler.layer_names:
                raise ValueError("feature_layer_name \'{}\' is not found in source_model".format(layer_name))

        self._feature_layer_names = value

    @property
    def source_model(self):
        """
        Model to extract features from.
        """
        return self._source_model

    @source_model.setter
    def source_model(self, value):
        if not isinstance(value, mx.mod.Module):
            error = ("source_model must be a valid `mxnet.mod.Module` object because"
                     " MetaModelRepurposer requires source_model to extract features for both repurpose and predict."
                     " Instead got type: {}".format(type(value)))
            raise TypeError(error)

        self._source_model = value

    def repurpose(self, train_iterator: mx.io.DataIter):
        """
        Train a meta-model using features extracted from training data through the source neural network.

        Set self.target_model to the trained meta-model.

        :param train_iterator: Training data iterator to use to extract features from source_model.
        :type train_iterator: :class: `mxnet.io.DataIter`
        """

        # Validate if repurpose call is valid
        self._validate_before_repurpose()

        # For given training data, extract features from source neural network
        train_data = self.get_features_from_source_model(train_iterator)

        # Train target model using the extracted features (transfer and learn)
        self.target_model = self._train_model_from_features(train_data.features, train_data.labels,
                                                            train_data.feature_indices_per_layer)

        # Save data_shape and label_shapes
        self.provide_data = train_iterator.provide_data
        self.provide_label = train_iterator.provide_label

    def predict_probability(self, test_iterator: mx.io.DataIter):
        """
        Predict class probabilities on test data using the target_model i.e. repurposed meta-model.

        :param test_iterator: Test data iterator to return predictions for.
        :type test_iterator: mxnet.io.DataIter
        :return: Predicted probabilities.
        :rtype: :class:`numpy.ndarray`
        """

        # Validate if predict call is valid
        self._validate_before_predict()

        # For given test data, extract features from source neural network
        test_data = self.get_features_from_source_model(test_iterator)

        # Predict and return probabilities
        return self._predict_probability_from_features(test_data.features)

    def predict_label(self, test_iterator: mx.io.DataIter):
        """
        Predict class labels on test data using the target_model i.e. repurposed meta-model.

        :param test_iterator: Test data iterator to return predictions for.
        :type test_iterator: mxnet.io.DataIter
        :return: Predicted labels.
        :rtype: :class:`numpy.ndarray`
        """
        self._validate_before_predict()

        # For given test data, extract features from source neural network
        test_data = self.get_features_from_source_model(test_iterator)

        # Predict and return labels
        return self._predict_label_from_features(test_data.features)

    def get_features_from_source_model(self, data_iterator: mx.io.DataIter):
        """
        Extract feature outputs from feature_layer_names in source_model, merge and return all features and labels.

        In addition, return mapping of feature_layer_name to indices in feature array.

        :param data_iterator: Iterator for data to be passed through the source network and extract features.
        :type data_iterator: :class:`mxnet.io.DataIter`
        :return: features, feature_indices_per_layer and labels.
        :rtype: :class:`MetaModelData`
        """

        feature_dict, labels = self.source_model_handler.get_layer_output(data_iterator, self.feature_layer_names)

        # Merge feature outputs from all layers and save feature indices per layer
        # e.g. output of feature_dict {layer1:[f1,f2,f3], layer2:[f4,f5]} would be
        # [f1,f2,f3,f4,f5] and {layer1:[0,1,2], layer2:[3,4]}
        features = None
        feature_indices_per_layer = OrderedDict()  # mapping for {layer name, feature indices}
        for layer_name in feature_dict:
            if features is None:
                next_feature_index = 0
                features = feature_dict[layer_name]
            else:
                next_feature_index = features.shape[1]  # New feature column(axis=1) will be added at the end
                features = np.hstack([features, feature_dict[layer_name]])
            total_num_features = features.shape[1]
            feature_indices_per_layer[layer_name] = np.arange(next_feature_index, total_num_features)

        meta_model_data = MetaModelData(features, feature_indices_per_layer, labels)
        return meta_model_data

    @abstractmethod
    def _train_model_from_features(self, features, labels, feature_indices_per_layer):
        """
        Abstract method.
        """
        pass

    @abstractmethod
    def _predict_probability_from_features(self, features):
        """
        Abstract method.
        """
        pass

    @abstractmethod
    def _predict_label_from_features(self, features):
        """
        Abstract method.
        """
        pass


class MetaModelData:
    def __init__(self, features, feature_indices_per_layer, labels):
        """
        Structured data used by meta model repurposer to pass information extracted from source model.

        :param features: Features extracted from source neural network
        :type features: :class:`numpy.ndarray`
        :param feature_indices_per_layer: Mapping of feature_layer_names to indices in features array i.e. {layer_name,
            feature_indices}
        :type feature_indices_per_layer: OrderedDict[str, :class:`numpy.ndarray`]
        :param labels: Labels for target dataset.
        :type labels: :class:`numpy.ndarray`
        """
        self.features = features
        self.feature_indices_per_layer = feature_indices_per_layer
        self.labels = labels



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
from .__version__ import __version__


class constants:
    VERSION = __version__


class repurposer_keys:
    PARAMS = 'parameters'
    TARGET_MODEL = 'target_model'
    ATTRS = 'attributes'
    MODEL = 'model'
    GPU = 'gpu'
    CPU = 'cpu'
    TUPLE = 'tuple'
    CONTEXT_FN = 'context_function'
    NUM_DEVICES = 'num_devices'
    PROVIDE_DATA = 'provide_data'
    PROVIDE_LABEL = 'provide_label'
    REPURPOSER_CLASS = 'repurposer_class'


class meta_model_repurposer_keys:
    C = 'c'
    N_JOBS = 'n_jobs'
    TOL = 'tol'
    FEATURE_LAYERS = 'feature_layer_names'
    KERNEL = 'kernel'
    GAMMA = 'gamma'
    PROB_ESTIMATES = 'enable_probability_estimates'


class neural_network_repurposer_keys:
    TRANSFER_LAYER_NAME = 'transfer_layer_name'
    TARGET_CLASS_COUNT = 'target_class_count'
    LEARNING_RATE = 'learning_rate'
    OPTIMIZER = 'optimizer'
    OPTIMIZER_PARAMS = 'optimizer_params'
    BATCH_SIZE = 'batch_size'
    NUM_EPOCHS = 'num_epochs'
    FIXED_LAYERS = 'fixed_layers'
    RANDOM_LAYERS = 'random_layers'
    NUM_LAYERS_TO_DROP = 'num_layers_to_drop'


class neural_network_repurposer_constants:
    DEFAULT_LEARNING_RATE = 0.001


class serialization_constants:
    JSON_SUFFIX = '.json'
    SOURCE_SUFFIX = '_source'
    LABEL_SUFFIX = '_label'
    POSTERIOR_SUFFIX = '_posterior'


class serialization_keys:
    FILE_PATH = 'file_path'
    VERSION = 'version'
    LAST_LAYER_NAME_SOURCE = 'last_layer_name'
    LAST_LAYER_NAME_TARGET = 'last_layer_name_target'


class bnn_repurposer_keys:
    SIGMA = 'sigma'
    NUM_LAYERS = 'num_layers'
    N_HIDDEN = 'n_hidden'
    NUM_SAMPLES_MC = 'num_samples_mc'
    START_ANNEALING = 'start_annealing'
    END_ANNEALING = 'end_annealing'
    ANNEALING_WEIGHT = 'annealing_weight'
    DIM_INPUT = 'dim_input'
    NUM_CLASSES = 'num_classes'
    LEARNING_RATE = 'learning_rate'
    BATCH_SIZE = 'batch_size'
    NUM_EPOCHS = 'num_epochs'
    VERBOSE = 'verbose'
    BNN_CONTEXT_FUNCTION = 'bnn_context_function'
    NUM_SAMPLES_MC_PREDICT = 'num_samples_mc_prediction'


class bnn_constants:
    MEAN_INIT_POSTERIOR = 0.1
    SIGMA_INIT_POSTERIOR = 0.05
    BNN_OPTIMIZER = 'adam'


class gp_repurposer_keys:
    MAX_FUNCTION_EVALUATIONS = 'max_function_evaluations'
    APPLY_L2_NORM = 'apply_l2_norm'
    FEATURE_MEAN = 'feature_mean'
    OPTIMIZER = 'optimizer'
    NUM_INDUCING_SPARSE_GP = 'num_inducing_sparse_gp'
    L2_NORMALIZER = 'l2_normalizer'



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet.ndarray as nd

from .prob_base import Probability
from .utils import log_gaussian


class Prior(Probability):
    def __init__(self, shapes, ctx):
        super(Prior, self).__init__(shapes, ctx)


class GaussianPrior(Prior):

    """
    Gaussian Prior (iid) over a set of variables [W_1, ..., W_n] where W_i is an mxnet.ndarray

    :param params: Dictionary containing the parameters that characterize the probability distribution
      , i.e. mean and variance of each W_i.
    :type params: dict(str, :class:`mxnet.gluon.Parameter`)
    :param raw_params: Dictionary containing the values of the parameters stored in params
    :type raw_params: dict(str, list[:class:`mxnet.ndarray`]}
    :param shapes: Shape of the variables over which the probability is defined.
    :type shapes: list(tuple(int, int))
    :param ctx: MXNet context
    :type ctx: :class:`mxnet.Context`
    """

    def __init__(self, means, sigmas, shapes, ctx=None, fix_means=True, fix_sigmas=True):
        """
        Default constructor

        :param means: Mean of the prior. Can be:
                [np.array]: Mean shared across all variables.
                [np.array_1, ..., np.array_n]: Different mean for each variable W_i
        :type means: list[:class:`np.ndarray`]
        :param sigmas: Variance of the prior. Can be:
                [np.array]: Variance shared across all variables.
                [np.array_1, ..., np.array_n]: Different variance for each variable W_i
        :type sigmas: list[:class:`np.ndarray`]
        :param shapes: Define the shape of each variable W_i.
        :type shapes: list(tuple(int, int))
        :param fix_means: Fix the mean of the Prior.
        :type fix_means: boolean or list(boolean)
        :param fix_sigmas: Fix the variance of the Prior
        :type fix_sigmas: boolean or list(boolean)
        """

        super(GaussianPrior, self).__init__(shapes, ctx)
        self._register_params(means, sigmas, fix_means, fix_sigmas)

    def _register_params(self, means, sigmas, fix_means, fix_sigmas):
        self._register_param_value("mean", means, fix=fix_means)
        self._register_param_value("sigma", sigmas, fix=fix_sigmas)

    def log_pdf(self, obs):
        self.check_observation_shapes(obs)
        raw_params_ext = self._replicate_shared_parameters()
        return sum([nd.sum(log_gaussian(obs[ii], raw_params_ext["mean"][ii], raw_params_ext["sigma"][ii]))
                    for ii in range(len(self.shapes))])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
# matplotlib is loaded in GPy initialization and if a backend('Agg') is not set, loading fails in mac osx
import matplotlib
matplotlib.use('Agg')

# import repurposer base classes
from .repurposer import Repurposer  # noqa: F401
from .meta_model_repurposer import MetaModelRepurposer  # noqa: F401
from .neural_network_repurposer import NeuralNetworkRepurposer  # noqa: F401

# import repurposers
from .svm_repurposer import SvmRepurposer  # noqa: F401
from .lr_repurposer import LrRepurposer  # noqa: F401
from .bnn_repurposer import BnnRepurposer  # noqa: F401
from .gp_repurposer import GpRepurposer  # noqa: F401
from .neural_network_fine_tune_repurposer import NeuralNetworkFineTuneRepurposer  # noqa: F401
from .neural_network_random_freeze_repurposer import NeuralNetworkRandomFreezeRepurposer  # noqa: F401

# import load function
from .load_repurposer import load  # noqa: F401

# import version
from .__version__ import __version__  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx
from sklearn.svm import SVC

from .meta_model_repurposer import MetaModelRepurposer
from .constants import repurposer_keys
from .constants import meta_model_repurposer_keys as keys
from . import utils


class SvmRepurposer(MetaModelRepurposer):
    """
    Perform Transfer Learning through a Support Vector Machine (SVM) meta-model which repurposes the source neural
    network.

    :param source_model: Source neural network to do transfer learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param feature_layer_names: Name of layer(s) in source_model from which features should be transferred.
    :type feature_layer_names: list[str]
    :param context_function: MXNet context function that provides device type context.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use to extract features from source_model.
    :param float c: Penalty parameter C of the error term.
    :param string kernel: Specifies the kernel type to be used in the SVM algorithm in sklearn library. It must
                          be one of 'linear', 'poly', 'rbf', 'sigmoid', 'precomputed' or a callable.
    :param float gamma: Kernel coefficient for 'rbf', 'poly' and 'sigmoid'. If gamma is 'auto' then 1/n_features will
                        be used instead.
    :param bool enable_probability_estimates: Whether to enable probability estimates.
                                              This must be enabled for predict_probability to work and will slow down
                                              training.
    """
    def __init__(self, source_model: mx.mod.Module, feature_layer_names, context_function=mx.cpu, num_devices=1,
                 c=1.0, kernel='linear', gamma='auto', enable_probability_estimates=False):
        # Call base class constructor with parameters required for meta-models
        super(SvmRepurposer, self).__init__(source_model, feature_layer_names, context_function, num_devices)

        # Initialize SVM specific parameters
        self.c = c
        self.kernel = kernel
        self.gamma = gamma
        self.enable_probability_estimates = enable_probability_estimates

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor.

        :rtype: dict
        """
        param_dict = super(SvmRepurposer, self).get_params()
        param_dict[keys.C] = self.c
        param_dict[keys.KERNEL] = self.kernel
        param_dict[keys.GAMMA] = self.gamma
        param_dict[keys.PROB_ESTIMATES] = self.enable_probability_estimates
        return param_dict

    def _train_model_from_features(self, features, labels, feature_indices_per_layer=None):
        """
        Train an SVM model using features extracted from source neural network.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :param labels: Labels to use for training.
        :type labels: :class:`numpy.ndarray`
        :param feature_indices_per_layer: Mapping of feature_layer_names to indices in features array
                                          i.e. {layer_name, feature_indices} Note that this param is currently not
                                          consumed by svm_repurposer.
        :type feature_indices_per_layer: OrderedDict[str, :class:`numpy.ndarray`]
        :return: SVM model trained with given features and labels using sci-kit learn library.
        :rtype: :class: `sklearn.svm.SVC`
        """

        svm_classifier = SVC(C=self.c,
                             kernel=self.kernel,
                             gamma=self.gamma,
                             decision_function_shape='ovr',
                             random_state=1,
                             probability=self.enable_probability_estimates)
        svm_classifier.fit(features, labels)
        return svm_classifier

    def _predict_probability_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use sklearn's SVM predict_proba method and return predicted probabilities.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted probabilities.
        :rtype: :class:`numpy.ndarray`
        """
        if not self.target_model.probability:
            raise ValueError("Probability estimates should have been enabled during model training for this method to \
                             work")
        return self.target_model.predict_proba(features)

    def _predict_label_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use sklearn's SVM predict method and return predicted labels.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted labels.
        :rtype: :class:`numpy.ndarray`
        """
        return self.target_model.predict(features)

    def serialize(self, file_prefix):
        """
        Saves repurposer (excluding source model) to file_prefix.json.

        :param str file_prefix: Prefix to save file with.
        """
        output_dict = {}
        output_dict[repurposer_keys.PARAMS] = self.get_params()
        output_dict[repurposer_keys.TARGET_MODEL] = utils.sklearn_model_to_dict(self.target_model)
        output_dict.update(self._get_attributes())

        utils.save_json(file_prefix, output_dict)

    def deserialize(self, input_dict):
        """
        Uses dictionary to set attributes of repurposer.

        :param dict input_dict: Dictionary containing values for attributes to be set to.
        """
        self._set_attributes(input_dict)  # Set attributes of the repurposer from input_dict
        self.target_model = utils.sklearn_model_from_dict(SVC, input_dict[repurposer_keys.TARGET_MODEL])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
from .prob_base import Probability  # noqa: F401
from .prior import Prior, GaussianPrior  # noqa: F401
from .obs import Likelihood, Categorical  # noqa: F401
from .var import VariationalPosterior, GaussianVariationalPosterior  # noqa: F401
from .var_loss import BNNLoss  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import numpy as np
from mxnet import gluon
import mxnet as mx
from sklearn.preprocessing import Normalizer

from .prob import VariationalPosterior
from .prob.utils import replace_params_net


class BnnClassifier(object):
    """
    Repurpose neural network to create BNN meta-model through Transfer Learning

    :param model: neural network model
    :type model: :class:`mxnet.gluon.nn.Sequential`
    :param var_posterior: Variational posterior distribution
    :type var_posterior: :class:`xfer.prob.VariationalPosterior`
    :param normalizer: It is applied to the input features before computing the predictions
    :type normalizer: :class:`sklearn.preprocessing.Normalizer`
    """

    def __init__(self, model: gluon.nn.Sequential, var_posterior: VariationalPosterior,
                 normalizer: Normalizer):

        self.model = model
        self.var_posterior = var_posterior
        self.normalizer = normalizer

    def predict(self, features, num_samples_mc_prediction=100, context=mx.cpu()):
        """
        Function to make predictions

        :param features: Features extracted from source neural network
        :type features: :class:`numpy.ndarray`
        :param int num_samples_mc_prediction: Num of samples from the posterior to compute the predictions
        :param context: MXNet context
        :type context: :class:`mxnet.Context`
        :return: Tuple containing predicted labels and probabilities
        :rtype: tuple(:class:`numpy.ndarray`, :class:`numpy.ndarray`)
        """
        if self.var_posterior.get_mean() is None:
            raise RuntimeError("Repurposer is not trained!")

        n_te, dim = features.shape
        features = features.astype(np.dtype(np.float32))
        features = self.normalizer.transform(features)
        data = mx.nd.array(features)
        data = data.as_in_context(context).reshape((-1, dim))

        for j in range(num_samples_mc_prediction):
            layer_params = self.var_posterior.generate_sample()

            replace_params_net(layer_params, self.model, context)

            output_temp = self.model(data)
            output = mx.nd.softmax(output_temp).asnumpy()

            if j == 0:
                predictions = np.zeros((n_te, output.shape[1]))

            for i in range(n_te):
                p = output[i, :].astype(np.dtype(np.float64))
                p /= p.sum()
                predictions[i, :] += np.random.multinomial(1, p)

        return predictions.argmax(axis=1), predictions/float(num_samples_mc_prediction)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx
import numpy as np
import GPy
from sklearn.preprocessing import LabelBinarizer, Normalizer

from .meta_model_repurposer import MetaModelRepurposer
from .constants import gp_repurposer_keys as keys
from .constants import repurposer_keys
from . import utils


class GpRepurposer(MetaModelRepurposer):
    """
    Repurpose source neural network to create a Gaussian Process (GP) meta-model through Transfer Learning.

    :param source_model: Source neural network to do transfer learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param feature_layer_names: Name of layer(s) in source_model from which features should be transferred.
    :type feature_layer_names: list[str]
    :param context_function: MXNet context function that provides device type context.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use to extract features from source_model.
    :param int max_function_evaluations: Maximum number of function evaluations to perform in GP optimization.
    :param bool apply_l2_norm: Whether to apply L2 normalization after extracting features from source neural network.
                               If set to True, L2 normalization will be applied to features before passing to GP during
                               training and prediction.
    """
    def __init__(self, source_model: mx.mod.Module, feature_layer_names, context_function=mx.context.cpu, num_devices=1,
                 max_function_evaluations=100, apply_l2_norm=False):
        # Call base class constructor with parameters required for meta-models
        super().__init__(source_model, feature_layer_names, context_function, num_devices)
        self.max_function_evaluations = max_function_evaluations
        self.apply_l2_norm = apply_l2_norm

        # Mean of features to use for normalization. Computed in training phase.
        # Used to normalize features in training and in prediction.
        self.feature_mean = None

        # Optimizer to use for training GP model
        self.optimizer = 'lbfgs'

        # Number of inducing points to use for sparse GP
        self.NUM_INDUCING_SPARSE_GP = 100

        # Normalizer to use when apply_l2_norm flag is set
        self.l2_normalizer = Normalizer(norm='l2')

    def _train_model_from_features(self, features, labels, feature_indices_per_layer):
        """
        Train GP classification models using features extracted from the source neural network.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :param labels: Labels to use for training.
        :type labels: :class:`numpy.ndarray`
        :param feature_indices_per_layer: Mapping of feature_layer_names to dimension indices in features array
                                          i.e. {layer_name, feature_indices}.
                                          Used to build separate kernels for features from different layers.
        :type feature_indices_per_layer: OrderedDict[str, :class:`numpy.ndarray`]
        :return: List of GP classification models trained for each class (one-vs-all) in given training data.
                 List of Sparse GP models returned if number of training instances is greater than
                 NUM_INDUCING_SPARSE_GP.
                 If there are only two classes in training data, then the output list contains a single model.
        :rtype: list[:class:`GPy.models.GPClassification`] or list[:class:`GPy.models.SparseGPClassification`]
        """
        # Normalize features to train on
        self.feature_mean = features.mean(axis=0)  # Compute mean for each feature across all training instances
        normalized_features = features - self.feature_mean  # Normalize features to have zero mean
        if self.apply_l2_norm:  # Apply L2 normalization if flag is set
            normalized_features = self.l2_normalizer.fit_transform(normalized_features)

        # Binarize labels in a one-vs-all fashion to train a separate model for each class
        # Output contains 'c' columns (c=number of classes) and each column contains binary labels w.r.t to that class.
        # If number of classes is two, then the output contains a single column with values 0 and 1
        binarized_labels = LabelBinarizer().fit_transform(labels)

        # Build kernel using given feature indices
        kernel = self._build_kernel(feature_indices_per_layer)

        # Do spare GP if number of training instances is greater than chosen number of inducing points
        # Otherwise, do normal GP classification because the data set is small
        num_training_instances = features.shape[0]
        do_sparse_gp = (num_training_instances > self.NUM_INDUCING_SPARSE_GP)

        # Train a GP model for each class (one-vs-all) if there are more than two classes, and one model if there
        # are only two classes
        num_models = binarized_labels.shape[1]
        models = [None] * num_models
        for model_index in range(num_models):
            binary_labels_for_current_class = binarized_labels[:, model_index:model_index+1]
            input_kernel = kernel.copy()  # Pass copy of kernel to avoid original kernel being updated by GPy
            models[model_index] = self._train_model_for_binary_label(binary_label=binary_labels_for_current_class,
                                                                     features=normalized_features,
                                                                     kernel=input_kernel,
                                                                     do_sparse_gp=do_sparse_gp)
        return models

    def _train_model_for_binary_label(self, features, binary_label, kernel, do_sparse_gp):
        # Train a GPy model for binary classification with given features and kernel
        if do_sparse_gp:
            model = GPy.models.SparseGPClassification(X=features, Y=binary_label, kernel=kernel,
                                                      num_inducing=self.NUM_INDUCING_SPARSE_GP)
        else:
            model = GPy.models.GPClassification(X=features, Y=binary_label, kernel=kernel)
        model.optimize(optimizer=self.optimizer, max_iters=self.max_function_evaluations)
        return model

    @staticmethod
    def _build_kernel(feature_indices_per_layer):
        """
        Build separate RBF kernels for features from different layers and return the kernel that results from adding all
        feature specific kernels.

        :param feature_indices_per_layer: Mapping of feature_layer_names to dimension indices in features array
                                          i.e. {layer_name, feature_indices}.
        :type feature_indices_per_layer: dict[str, :class:`numpy.ndarray`]
        :return: GPy RBF kernel if all features are from single layer or GPy Add kernel if features are from multiple
                 layers.
        :rtype: :class:`GPy.kern.RBF` or :class:`GPy.kern.Add`
        """
        all_kernels = None
        for layer_name in feature_indices_per_layer:
            active_dims = feature_indices_per_layer[layer_name]  # feature indices corresponding to current layer
            kernel = GPy.kern.RBF(input_dim=active_dims.size, name=layer_name, active_dims=active_dims.tolist())
            if all_kernels is None:
                all_kernels = kernel
            else:
                all_kernels += kernel
        return all_kernels

    def _predict_probability_from_features(self, features):
        """
        Compute predictions using self.target_model with features extracted from source neural network.
        self.target_model is a list of GP classification models trained for each class in a one-vs-all fashion.
        Use GPy's predict method on each model and compute probabilities predicted for individual classes.
        The individual class probabilities are then normalized such that their sum is 1.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Normalized one-vs-all probabilities.
        :rtype: :class:`numpy.ndarray`
        """
        normalized_features = features - self.feature_mean  # Normalize features using the training features' means
        if self.apply_l2_norm:  # Apply L2 normalization if flag is set
            normalized_features = self.l2_normalizer.transform(normalized_features)

        num_gp_models = len(self.target_model)
        if num_gp_models == 1:
            # When there are only two classes, get probability for class_1 (P) and calculate probability for class_0
            # (1-P)
            prediction, _ = self.target_model[0].predict(normalized_features)
            normalized_predictions = np.hstack([1.0-prediction, prediction])

        else:
            # When there are more than two classes, get one-vs-all prediction scores for each class
            # from binary GP models
            predictions_per_class = []
            for model_id in range(num_gp_models):
                binary_gp_model = self.target_model[model_id]
                binary_prediction, _ = binary_gp_model.predict(normalized_features)
                predictions_per_class.append(binary_prediction)

            # Convert scores list to numpy array
            predictions = np.nan_to_num(np.hstack(predictions_per_class))

            # Normalize individual predictions to sum up to 1
            sum_of_predictions_per_instance = np.sum(predictions, axis=1).reshape(predictions.shape[0], 1)
            normalized_predictions = predictions / sum_of_predictions_per_instance

        return normalized_predictions

    def _predict_label_from_features(self, features):
        """
        Return labels predicted using target_model with features extracted from source neural network.
        target_model is a list of GP classification models trained for each class in a one-vs-all fashion.
        Using GPy's predict method, compute one-vs-all probabilities per class and return the class with
        maximum probability.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted labels i.e. Class with maximum probability for each test instance.
        :rtype: :class:`numpy.ndarray`
        """
        predictions = self._predict_probability_from_features(features)
        labels = np.argmax(predictions, axis=1)  # Select label with maximum prediction for each test instance
        return labels

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor's argument list.

        :rtype: dict
        """
        params_dict = super().get_params()
        params_dict[keys.MAX_FUNCTION_EVALUATIONS] = self.max_function_evaluations
        params_dict[keys.APPLY_L2_NORM] = self.apply_l2_norm
        return params_dict

    def _get_attributes(self):
        """
        Get parameters of repurposer not in constructor's argument list.

        :rtype: dict
        """
        attributes_dict = super()._get_attributes()
        attributes_dict[keys.OPTIMIZER] = self.optimizer
        attributes_dict[keys.NUM_INDUCING_SPARSE_GP] = self.NUM_INDUCING_SPARSE_GP
        attributes_dict[keys.FEATURE_MEAN] = self.feature_mean.tolist()
        return attributes_dict

    def _set_attributes(self, input_dict):
        """
        Set attributes of class from input_dict.
        These attributes are the same as those returned by get_attributes method.

        :param input_dict: Dictionary containing attribute values.
        :return: None
        """
        super()._set_attributes(input_dict)
        self.optimizer = input_dict[keys.OPTIMIZER]
        self.NUM_INDUCING_SPARSE_GP = input_dict[keys.NUM_INDUCING_SPARSE_GP]
        self.feature_mean = np.array(input_dict[keys.FEATURE_MEAN])

    def serialize(self, file_prefix):
        """
        Serialize the GP repurposer (model and supporting info) and save to file.

        :param str file_prefix: Prefix of file path to save the serialized repurposer to.
        :return: None
        """
        # Get constructor params. This will be used to recreate repurposer object in deserialization flow.
        output_dict = {repurposer_keys.PARAMS: self.get_params()}

        # Get rest of the attributes to save with the repurposer.
        output_dict.update(self._get_attributes())

        # Serialize the GP models and save to output dictionary
        output_dict[repurposer_keys.TARGET_MODEL] = self._serialize_target_gp_models(save_data=True)

        # Save the serialized params and attributes to file
        utils.save_json(file_prefix, output_dict)

    def deserialize(self, input_dict):
        """
        Uses dictionary to set attributes of repurposer.

        :param dict input_dict: Dictionary containing values for attributes to be set to.
        :return: None
        """
        # Set attributes of the repurposer from input_dict
        self._set_attributes(input_dict)

        # Deserialize and set the target GP models
        self.target_model = self._deserialize_target_gp_models(input_dict[repurposer_keys.TARGET_MODEL])

    def _serialize_target_gp_models(self, save_data=True):
        # Serialize the gp models trained per class and add to output list
        serialized_models = []
        for model in self.target_model:
            serialized_models.append(model.to_dict(save_data=save_data))
        return serialized_models

    @staticmethod
    def _deserialize_target_gp_models(serialized_target_model):
        # Deserialize the GP models trained per class and return
        deserialized_models = []
        for model_dict in serialized_target_model:
            deserialized_models.append(GPy.core.Model.from_dict(model_dict))
        return deserialized_models



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
"""
Layer classes for adding new layers to models
"""

import ast
import mxnet as mx

from . import consts


def _parse_bool_string(string):
    return string.lower() in ['true', '1']


def _get_input_symbol(layer_dict):
    input_id = layer_dict[consts.INPUTS][0][0] + layer_dict[consts.ELEMENTS_OFFSET]
    input_name = layer_dict[consts.ID2NAME][input_id]
    input_symbol = layer_dict[consts.PREV_SYMBOLS][input_name]
    return input_symbol


class LayerFactory(object):
    """
    Base class for neural network layer.

    :param str name: Name of the resulting layer.
    """
    def __init__(self, name):
        self.attributes = {consts.NAME: name}
        self.output = False
        self.SymbolClass = None
        self.layer_type = None

    def create_layer(self, layer_input):
        """
        Add layer to symbol graph provided in layer_input.

        :param layer_input: Symbol to add layer to.
        :type layer_input: :class:`mx.symbol.Symbol`
        :rtype: :class:`mx.symbol.Symbol`
        """
        return self.SymbolClass(data=layer_input, **self.attributes)

    @staticmethod
    def _from_dict(layer_dict):
        """
        Function to return layer object from dictionary.

        :rtype: :class:`LayerFactory`
        :raises ValueError: If layer type is not supported by package.
        """
        # Table of layer names to layer classes
        class_dict = {
            'FullyConnected': FullyConnected,
            'Convolution': Convolution,
            'Activation': Activation,
            'Pooling': Pooling,
            'Flatten': Flatten,
            'SoftmaxOutput': SoftmaxOutput,
            'Dropout': Dropout,
            'Concat': Concat,
            'BatchNorm': BatchNorm,
            '_Plus': Add,
            'SVMOutput': SVMOutput,
            'Embedding': Embedding,
            'mean': Mean,
            'null': Data
        }
        try:
            constructor_function = class_dict[layer_dict[consts.OPERATION]]._from_dict
        except KeyError:
            raise ValueError('Unsupported layer type found in dict: ' + layer_dict[consts.OPERATION])
        return constructor_function(layer_dict)


class FullyConnected(LayerFactory):
    """
    Class for fully connected layers.

    :param str name: Name of the resulting layer.
    :param int num_hidden: Number of hidden nodes of the output.
    :param boolean no_bias: Whether to disable bias parameter.
    """
    def __init__(self, name, num_hidden, no_bias=False):
        super(FullyConnected, self).__init__(name=name)
        self.attributes['num_hidden'] = num_hidden
        self.attributes['no_bias'] = no_bias
        self.layer_type = consts.LayerType.FULLYCONNECTED.value
        self.SymbolClass = mx.sym.FullyConnected

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        num_hidden = int(layer_dict[consts.ATTRIBUTES]['num_hidden'])
        no_bias = layer_dict[consts.ATTRIBUTES].get('no_bias', 'False') == 'True'
        input_symbol = _get_input_symbol(layer_dict)
        return FullyConnected(name=name, num_hidden=num_hidden, no_bias=no_bias), input_symbol


class SoftmaxOutput(LayerFactory):
    """Class for softmax output layers.

    :param str name: Name of the resulting layer.
    """
    def __init__(self, name):
        super(SoftmaxOutput, self).__init__(name=name)
        self.layer_type = consts.LayerType.SOFTMAXOUTPUT.value
        self.SymbolClass = mx.sym.SoftmaxOutput
        self.output = True

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        input_symbol = _get_input_symbol(layer_dict)
        return SoftmaxOutput(name=name), input_symbol


class Activation(LayerFactory):
    """Class for activation layers.

    :param name: Name of the resulting layer.
    :param str act_type: Activation function to be applied. Can be selected among:

        - 'relu'
        - 'sigmoid'
        - 'softrelu'
        - 'tanh'
    """
    def __init__(self, name, act_type='relu'):
        super(Activation, self).__init__(name=name)
        self.attributes['act_type'] = act_type
        self.layer_type = consts.LayerType.ACTIVATION.value
        self.SymbolClass = mx.sym.Activation

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        act_type = layer_dict[consts.ATTRIBUTES]['act_type']
        input_symbol = _get_input_symbol(layer_dict)
        return Activation(name=name, act_type=act_type), input_symbol


class Convolution(LayerFactory):
    """Class for convolution layers.

    :param str name: Name of the resulting layer.
    :param tuple kernel: Convolution kernel size.
    :param int num_filter: Convolution filter (channel) number.
    :param tuple dilate: Convolution dilate.
    :param tuple pad: Pad for convolution.
    :param str cudnn_tune: Whether to pick convolution algorithm by running performance test. Can be selected among:

        - 'None'
        - 'fastest'
        - 'limited_workspace'
        - 'off'
    :param tuple stride: Convolution stride.
    :param boolean no_bias: Whether to disable bias parameter.
    :param boolean cudnn_off: Turn off cudnn for this layer.
    :param layout: Set layout for input, output and weight. Empty for default layout: NCW for 1d, NCHW for 2d and NCDHW
                   for 3d. Can be selected among:

                   - 'None'
                   - 'NCDHW'
                   - 'NCHW'
                   - 'NCW'
                   - 'NDHWC'
                   - 'NHWC'
    :param int num_group: Number of group partitions (Non-negative).
    """
    def __init__(self, name, kernel, num_filter, dilate=(), pad=(), cudnn_tune=None, stride=(), no_bias=False,
                 cudnn_off=False, layout=None, num_group=1):
        super(Convolution, self).__init__(name=name)
        self.attributes['kernel'] = kernel
        self.attributes['num_filter'] = num_filter
        self.attributes['dilate'] = dilate
        self.attributes['pad'] = pad
        self.attributes['cudnn_tune'] = cudnn_tune
        self.attributes['stride'] = stride
        self.attributes['no_bias'] = no_bias
        self.attributes['cudnn_off'] = cudnn_off
        self.attributes['layout'] = layout
        self.attributes['num_group'] = num_group
        self.layer_type = consts.LayerType.CONVOLUTION.value
        self.SymbolClass = mx.sym.Convolution

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        kernel = ast.literal_eval(layer_dict[consts.ATTRIBUTES]['kernel'])
        pad = ast.literal_eval(layer_dict[consts.ATTRIBUTES].get('pad', '()'))
        stride = ast.literal_eval(layer_dict[consts.ATTRIBUTES].get('stride', '()'))
        dilate = ast.literal_eval(layer_dict[consts.ATTRIBUTES].get('dilate', '()'))
        num_filter = int(layer_dict[consts.ATTRIBUTES]['num_filter'])
        cudnn_tune = layer_dict[consts.ATTRIBUTES].get('cudnn_tune', 'None')
        no_bias = layer_dict[consts.ATTRIBUTES].get('no_bias', 'False') == 'True'
        cudnn_off = _parse_bool_string(layer_dict[consts.ATTRIBUTES].get('cudnn_off', 'False'))
        layout = layer_dict[consts.ATTRIBUTES].get('layout', None)
        num_group = int(layer_dict[consts.ATTRIBUTES].get('num_group', 1))
        input_symbol = _get_input_symbol(layer_dict)
        return Convolution(name=name, kernel=kernel, num_group=num_group, num_filter=num_filter, dilate=dilate,
                           pad=pad, cudnn_tune=cudnn_tune, stride=stride, no_bias=no_bias,
                           cudnn_off=cudnn_off, layout=layout), input_symbol


class Data(LayerFactory):
    """Class for data layer.

    :param name: Name of the resulting layer.
    """
    def __init__(self, name):
        super(Data, self).__init__(name=name)
        self.layer_type = consts.LayerType.DATA.value

    def create_layer(self, layer_input=None):
        """
        Returns variable symbol with name 'data'.

        :param layer_input: Symbol to add layer to.
        :type layer_input: :class:`mx.symbol.Symbol`
        """
        return mx.sym.var(**self.attributes)

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        data_name = layer_dict[consts.DATA]
        if name == data_name:
            return Data(name=name), None
        else:
            # nodes that have 'op' as 'null' will return None, None
            return None, None


class Pooling(LayerFactory):
    """Class for pooling layer.

    :param str name: Name of the resulting layer.
    :param str pool_type: Pooling type to be applied. Can be selected among:

        - 'avg'
        - 'max'
        - 'sum'
    :param tuple kernel: Pooling kernel size.
    :param tuple pad: Pad for pooling.
    :param str pooling_convention: Pooling convention to be applied. Can be selected among:

        - 'valid' (default)
        - 'full'
    :param tuple stride: Stride for pooling.
    :param boolean global_pool: Ignore kernel size, do global pooling based on current input feature map.
    """
    def __init__(self, pool_type, name, kernel=(), pad=(), pooling_convention='valid', stride=(),
                 global_pool=False):
        super(Pooling, self).__init__(name=name)
        self.attributes['kernel'] = kernel
        self.attributes['pool_type'] = pool_type
        self.attributes['pad'] = pad
        self.attributes['pooling_convention'] = pooling_convention
        self.attributes['stride'] = stride
        self.attributes['global_pool'] = global_pool
        self.layer_type = consts.LayerType.POOLING.value
        self.SymbolClass = mx.sym.Pooling

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        kernel = ast.literal_eval(layer_dict[consts.ATTRIBUTES]['kernel'])
        pad = ast.literal_eval(layer_dict[consts.ATTRIBUTES].get('pad', '()'))
        pool_type = layer_dict[consts.ATTRIBUTES]['pool_type']
        pooling_convention = layer_dict[consts.ATTRIBUTES].get('pooling_convention', 'valid')
        stride = ast.literal_eval(layer_dict[consts.ATTRIBUTES].get('stride', '()'))
        global_pool = layer_dict[consts.ATTRIBUTES].get('global_pool', 'False') == 'True'
        input_symbol = _get_input_symbol(layer_dict)
        return Pooling(name=name, kernel=kernel, pad=pad, pool_type=pool_type,
                       pooling_convention=pooling_convention, stride=stride, global_pool=global_pool), input_symbol


class Flatten(LayerFactory):
    """Class for flatten layer.

    :param str name: Name of the resulting layer.
    """
    def __init__(self, name):
        super(Flatten, self).__init__(name=name)
        self.layer_type = consts.LayerType.FLATTEN.value
        self.SymbolClass = mx.sym.flatten

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        input_symbol = _get_input_symbol(layer_dict)
        return Flatten(name=name), input_symbol


class Dropout(LayerFactory):
    """Class for dropout layer.

    :param str name: Name of the resulting layer.
    :param float p: Fraction of the input that gets dropped out during training time.
    """
    def __init__(self, name, p=0.5):
        super(Dropout, self).__init__(name=name)
        self.attributes['p'] = p
        self.layer_type = consts.LayerType.DROPOUT.value
        self.SymbolClass = mx.sym.Dropout

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        default_p = 0.5  # default dropout fraction
        p = float(layer_dict[consts.ATTRIBUTES].get('p', default_p))
        input_symbol = _get_input_symbol(layer_dict)
        return Dropout(name=name, p=p), input_symbol


class Concat(LayerFactory):
    """Class for concat layer.

    :param str name: Name of the resulting layer.
    :param int dim: The dimension to be concatenated.
    """
    def __init__(self, name, dim=1):
        super(Concat, self).__init__(name=name)
        self.attributes['dim'] = dim
        self.layer_type = consts.LayerType.CONCAT.value
        self.SymbolClass = mx.sym.concat

    def create_layer(self, layer_input):
        """
        Add layer to symbol graph provided in layer_input.

        :param layer_input: List of symbols to have their output arrays concatenated.
        :type layer_input: list[:class:`Symbol`]
        :rtype: :class:`mx.symbol.Symbol`
        """
        data = layer_input if type(layer_input) == list else [layer_input]
        return mx.sym.concat(*data, **self.attributes)

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        dim = int(layer_dict[consts.ATTRIBUTES]['dim'])
        input_symbol = []
        for input_tuple in layer_dict[consts.INPUTS][:-1]:
            input_id = input_tuple[0]
            input_name = layer_dict[consts.ID2NAME][input_id]
            input_symbol.append(layer_dict[consts.PREV_SYMBOLS][input_name])
        return Concat(name=name, dim=dim), input_symbol


class BatchNorm(LayerFactory):
    """Class for batch normalisation layer.

    :param str name: Name of the resulting layer.
    :param float eps: Epsilon to prevent div 0.
    :param boolean fix_gamma: Fix gamma while training.
    :param float momentum: Momentum for moving average.
    :param boolean use_global_stats: Whether to use global moving statistics instead of local batch-norm. This will
                                     force change batch-norm into a scale shift operator.
    """
    def __init__(self, name, eps=0.001, fix_gamma=True, momentum=0.9, use_global_stats=False):
        super(BatchNorm, self).__init__(name=name)
        self.attributes['eps'] = eps
        self.attributes['fix_gamma'] = fix_gamma
        self.attributes['momentum'] = momentum
        self.attributes['use_global_stats'] = use_global_stats
        self.layer_type = consts.LayerType.BATCHNORM.value
        self.SymbolClass = mx.sym.BatchNorm

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        eps = float(layer_dict[consts.ATTRIBUTES].get('eps', 0.001))
        momentum = float(layer_dict[consts.ATTRIBUTES].get('momentum', 0.9))
        fix_gamma = layer_dict[consts.ATTRIBUTES].get('fix_gamma', 'True') == 'True'
        use_global_stats = layer_dict[consts.ATTRIBUTES].get('use_global_stats', 'False') == 'True'
        input_symbol = _get_input_symbol(layer_dict)
        return BatchNorm(name=name, eps=eps, fix_gamma=fix_gamma, momentum=momentum,
                         use_global_stats=use_global_stats), input_symbol


class Add(LayerFactory):
    """Class for add layer.

    :param str name: Name of the resulting layer.
    """
    def __init__(self, name):
        super(Add, self).__init__(name=name)
        self.layer_type = consts.LayerType.PLUS.value

    def create_layer(self, layer_input):
        """
        Add layer to symbol graph provided in layer_input.

        :param layer_input: List of two symbols to perform addition on.
        :type layer_input: list[:class:`mx.symbol.Symbol`]
        """
        if len(layer_input) != 2:
            raise ValueError('layer_input must be a list of 2 Symbols'.format(len(layer_input)))
        return layer_input[0].__add__(layer_input[1])

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        input_symbol = _get_input_symbol(layer_dict)
        return Add(name=name), input_symbol


class SVMOutput(LayerFactory):
    """Class for SVM output layer.

    :param str name: Name of the resulting layer.
    :param label: Class label for the input data.
    :type label: :class:`Symbol`
    :param float margin: The loss function penalises outputs that lie outside this margin. Default margin is 1.
    :param float regularization_coefficient: Regularization parameter for the SVM. This balances the tradeoff between
                                             coefficient size and error.
    :param boolean use_linear: Whether to use L1-SVM objective. L2-SVM objective is used by default.
    """
    def __init__(self, name, label=None, margin=1, regularization_coefficient=1, use_linear=0):
        super(SVMOutput, self).__init__(name=name)
        self.attributes['label'] = label
        self.attributes['margin'] = margin
        self.attributes['regularization_coefficient'] = regularization_coefficient
        self.attributes['use_linear'] = use_linear
        self.layer_type = consts.LayerType.SVMOUTPUT.value
        self.output = True
        self.SymbolClass = mx.sym.SVMOutput

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        margin = float(layer_dict[consts.ATTRIBUTES].get('margin', 1))
        regularization_coefficient = float(layer_dict[consts.ATTRIBUTES].get('regularization_coefficient', 1))
        use_linear = layer_dict[consts.ATTRIBUTES]['use_linear'] == 'True' or \
            int(layer_dict[consts.ATTRIBUTES]['use_linear']) == 1
        input_symbol = _get_input_symbol(layer_dict)
        return SVMOutput(name=name, margin=margin, regularization_coefficient=regularization_coefficient,
                         use_linear=use_linear), input_symbol


class Embedding(LayerFactory):
    """Class for Embedding layer. Maps integer indices to vector representations (embeddings).

    :param str name: Name of the resulting layer.
    :param int input_dim: Vocabulary size of the input indices.
    :param int output_dim: Dimension of the embedding vectors.
    :param weight: The embedding weight matrix.
    :type weight: :class:`mx.symbol.Symbol`
    :param dtype: Data type of weight.
    :type dtype: {'float16', 'float32', 'float64', 'int32', 'int64', 'int8', 'uint8'}, default='float32'
    """
    def __init__(self, name, input_dim, output_dim, weight=None, dtype='float32'):
        super(Embedding, self).__init__(name=name)
        self.attributes['input_dim'] = input_dim
        self.attributes['output_dim'] = output_dim
        self.attributes['weight'] = weight
        self.attributes['dtype'] = dtype
        self.layer_type = consts.LayerType.EMBEDDING.value
        self.SymbolClass = mx.sym.Embedding

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        input_dim = int(layer_dict[consts.ATTRIBUTES]['input_dim'])
        output_dim = int(layer_dict[consts.ATTRIBUTES]['output_dim'])
        dtype = layer_dict[consts.ATTRIBUTES].get('dtype', 'float32')
        input_symbol = _get_input_symbol(layer_dict)
        return Embedding(name=name, input_dim=input_dim, output_dim=output_dim, dtype=dtype), input_symbol


class Mean(LayerFactory):
    """Class for mean layer. Computes the mean of array elements over given axis.

    :param str name: Name of the resulting layer.
    :param axis: The axis along which to perform the reduction.
    :type axis: Shape or None, optional, default=None
    :param boolean keepdims: If this is set to True, the reduced axis are left in the result as dimension with size one.
    :param boolean exclude: Whether to perform reduction on axis that are NOT in axis instead.
    """
    def __init__(self, name, axis=[], keepdims=0, exclude=0):
        super(Mean, self).__init__(name=name)
        self.attributes['axis'] = axis
        self.attributes['keepdims'] = keepdims
        self.attributes['exclude'] = exclude
        self.layer_type = consts.LayerType.MEAN.value
        self.SymbolClass = mx.sym.mean

    def _from_dict(layer_dict):
        name = layer_dict[consts.NAME]
        keepdims = _parse_bool_string(layer_dict[consts.ATTRIBUTES].get('keepdims', 'False'))
        exclude = _parse_bool_string(layer_dict[consts.ATTRIBUTES].get('exclude', 'False'))
        axis = eval(layer_dict[consts.ATTRIBUTES]['axis'])
        input_symbol = _get_input_symbol(layer_dict)
        return Mean(name=name, keepdims=keepdims, exclude=exclude, axis=axis), input_symbol



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
from mxnet import autograd, gluon
from sklearn.preprocessing import Normalizer
import logging

from .meta_model_repurposer import MetaModelRepurposer
from .prob import GaussianPrior, GaussianVariationalPosterior, BNNLoss, Categorical
from .bnn_classifier import BnnClassifier
from .prob.utils import replace_params_net
from .constants import repurposer_keys, serialization_constants, serialization_keys
from .constants import bnn_repurposer_keys as keys
from .constants import bnn_constants as consts

from . import utils


class BnnRepurposer(MetaModelRepurposer):
    """
    Perform Transfer Learning through a Bayesian Neural Network (BNN) meta-model which repurposes the source neural
    network.

    :param source_model: Source neural network to do transfer learning from.
    :type source_model: :class:`mxnet.mod.Module`
    :param feature_layer_names: Name of layer(s) in source_model from which features should be transferred.
    :type feature_layer_names: list[str]
    :param context_function: MXNet context function that provides device type context. It is used to extract features
        with the source model.
    :type context_function: function(int)->:class:`mx.context.Context`
    :param int num_devices: Number of devices to use to extract features from source_model.
    :param bnn_context_function: MXNet context function used to train the BNN.
    :type bnn_context_function: function(int)->:class:`mx.context.Context`
    :param float sigma: Standard deviation of the Gaussian prior used for the weights of the BNN meta-model
        (w_i \sim N(0, \sigma^2)).
    :param int num_layers: Number of layers of the BNN meta-model.
    :param int n_hidden: Dimensionality of the hidden layers of the BNN meta-model
        (all hidden layers have the same dimensionality).
    :param int num_samples_mc: Number of samples used for the Monte Carlo approximation of the variational bound.
    :param float learning_rate: Learning rate for the BNN meta-model training.
    :param int batch_size: Mini-batch size for the BNN meta-model training.
    :param int num_epochs: Number of epochs for the BNN meta-model training.
    :param int start_annealing: To help the training of the BNN meta-model, we anneal the KL term using a weight
        that varies fromm zero to one. start_annealing determines the epoch at which the annealing weight start
        to increase linearly until it reaches one in the epoch given by end_annealing.
    :param int end_annealing: Determines the epoch at which the annealing process of the KL term ends.
    :param float step_annealing_sample_weight: Amount that the annealing weight is incremented in each epoch
       (from start_annealing to end_annealing).
    :param int num_samples_mc_prediction: Number of Monte Carlo samples to use on prediction.
    :param bool verbose: Flag to control whether accuracy monitoring is logged during repurposing.
    :param float annealing_weight: Annealing weight in the current epoch.
    :param train_acc: Accuracy in training set in each epoch.
    :type train_acc: list[float]
    :param test_acc: Accuracy in validation set in each epoch.
    :type test_acc: list[float]
    :param moving_loss_total: Total loss (negative ELBO) smoothed across epochs.
    :type moving_loss_total: list[float]
    :param average_loss: Average loss (negative ELBO) per data point.
    :type average_loss: list[float]
    :param anneal_weights: Annealing weight used in each epoch.
    :type anneal_weights: list[float]
    """

    def __init__(self, source_model: mx.mod.Module, feature_layer_names, context_function=mx.cpu, num_devices=1,
                 bnn_context_function=mx.cpu, sigma=100.0, num_layers=1, n_hidden=10, num_samples_mc=3,
                 learning_rate=1e-3, batch_size=20, num_epochs=200, start_annealing=None, end_annealing=None,
                 num_samples_mc_prediction=100, verbose=0):

        # Call base class constructor with parameters required for meta-models
        super().__init__(source_model, feature_layer_names, context_function, num_devices)

        # Initialize BNN specific parameters
        self.sigma = sigma
        self.num_layers = num_layers
        self.n_hidden = n_hidden
        self.num_samples_mc = num_samples_mc
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.num_samples_mc_prediction = num_samples_mc_prediction
        self.verbose = verbose

        self.start_annealing = start_annealing
        self.end_annealing = end_annealing
        self.step_annealing_sample_weight = 1.0 / float(self.end_annealing - self.start_annealing)
        self.annealing_weight = 0.0

        # Initialize variables to track performance
        self.train_acc = []
        self.test_acc = []
        self.moving_loss_total = []
        self.current_loss_total = []
        self.average_loss = []
        self.anneal_weights = []

        # L2 normalization of the features
        self.normalizer = Normalizer(norm='l2')

        self.bnn_context_function = bnn_context_function
        self._context_bnn = self.bnn_context_function()

        # init parameters for constructing network to None. These will be set during repurposing
        self.dim_input = None
        self.num_classes = None

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor.

        :rtype: dict
        """
        param_dict = super().get_params()
        param_dict[keys.BNN_CONTEXT_FUNCTION] = utils.serialize_ctx_fn(self.bnn_context_function)
        param_dict[keys.SIGMA] = self.sigma
        param_dict[keys.NUM_LAYERS] = self.num_layers
        param_dict[keys.N_HIDDEN] = self.n_hidden
        param_dict[keys.NUM_SAMPLES_MC] = self.num_samples_mc
        param_dict[keys.LEARNING_RATE] = self.learning_rate
        param_dict[keys.BATCH_SIZE] = self.batch_size
        param_dict[keys.NUM_EPOCHS] = self.num_epochs
        param_dict[keys.START_ANNEALING] = self.start_annealing
        param_dict[keys.END_ANNEALING] = self.end_annealing
        param_dict[keys.VERBOSE] = self.verbose
        param_dict[keys.NUM_SAMPLES_MC_PREDICT] = self.num_samples_mc_prediction
        return param_dict

    def _get_attributes(self):
        """
        Get parameters of repurposer not in constructor.

        :rtype: dict
        """
        attr_dict = super()._get_attributes()
        attr_dict[keys.ANNEALING_WEIGHT] = self.annealing_weight
        attr_dict[keys.DIM_INPUT] = self.dim_input
        attr_dict[keys.NUM_CLASSES] = self.num_classes
        return attr_dict

    def _set_attributes(self, input_dict):
        """
        Set attributes of class from input_dict.
        These attributes are the same as those returned by get_attributes method.

        :param input_dict: Dictionary containing attribute values.
        :return: None
        """
        super()._set_attributes(input_dict)
        self.annealing_weight = input_dict[keys.ANNEALING_WEIGHT]
        self.dim_input = input_dict[keys.DIM_INPUT]
        self.num_classes = input_dict[keys.NUM_CLASSES]

    @property
    def start_annealing(self):
        """
        Determines the epoch at which the annealing process of the KL term starts.
        """
        return self._start_annealing

    @start_annealing.setter
    def start_annealing(self, value):
        if value is None:
            self._start_annealing = int(0.1*self.num_epochs)
        else:
            if value >= self.num_epochs:
                error = ("start_annealing must be smaller than the number of epochs."
                         " Instead got start_annealing = {}, number of epochs = {}.".format(value, self.num_epochs))
                raise ValueError(error)
            self._start_annealing = int(value)

    @property
    def end_annealing(self):
        """
        Determines the epoch at which the annealing process of the KL term ends.
        """
        return self._end_annealing

    @end_annealing.setter
    def end_annealing(self, value):
        if value is None:
            self._end_annealing = self.start_annealing + int(max(0.1 * self.num_epochs, 1))
        else:
            if value > self.num_epochs:
                error = ("end_annealing must be smaller than the number of epochs."
                         " Instead got end_annealing = {}, number of epochs = {}.".format(value, self.num_epochs))
                raise ValueError(error)

            if value <= self.start_annealing:
                error = ("end_annealing must be greater than start_annealing."
                         " Instead got start_annealing = {}, end_annealing = {}.".format(self.start_annealing, value))
                raise ValueError(error)

            self._end_annealing = int(value)

    def _build_nn(self, train_data, num_classes):
        _, (data, _) = next(enumerate(train_data))
        dim_input = data.shape[1]
        net = self._get_net(num_classes)
        shapes = self._get_shapes(net, data, dim_input)

        # save parameters used for constructing network
        self.dim_input = dim_input
        self.num_classes = num_classes

        return net, shapes

    def _get_net(self, num_classes):
        net = gluon.nn.Sequential()
        with net.name_scope():
            for i in range(self.num_layers):
                net.add(gluon.nn.Dense(self.n_hidden, activation="relu"))
            net.add(gluon.nn.Dense(num_classes))
        # Init parameters
        net.collect_params().initialize(mx.init.Xavier(magnitude=2.24), ctx=self._context_bnn)
        return net

    def _get_shapes(self, net, data, dim_input):
        # first pass to init size params
        data = data.as_in_context(self._context_bnn).reshape((-1, dim_input))
        net(data)
        shapes = [x.shape for x in net.collect_params().values()]
        return shapes

    def _update_annealing_weight(self, epoch):
        if (epoch > self.start_annealing) and (epoch <= self.end_annealing):
            self.annealing_weight += self.step_annealing_sample_weight

    def _compute_loss_grad(self, data, label_one_hot, num_train, net, bnn_loss):
        with autograd.record():
            # calculate the loss
            # Note: It triggers the hybrid_forward method of the bnn_loss object
            loss = bnn_loss(data, net, label_one_hot, self.num_samples_mc,
                            num_train, self.annealing_weight)
            # back-propagate for gradient calculation
            loss.backward()
        return loss

    def _evaluate_accuracy(self, data_iterator, net, layer_params):
        numerator = 0.
        denominator = 0.
        for i, (data, label) in enumerate(data_iterator):
            data = data.as_in_context(self._context_bnn).reshape((-1, data.shape[1]))
            label = label.as_in_context(self._context_bnn)
            replace_params_net(layer_params, net, self._context_bnn)
            output = net(data)
            predictions = nd.argmax(output, axis=1)
            numerator += nd.sum(predictions == label)
            denominator += data.shape[0]
        return (numerator / denominator).asscalar()

    def _monitor_acc(self, epoch, train_data, net, posterior_mean, test_data=None):
        """
        Monitor accuracy on training set and test set (if provided).
        """
        train_accuracy = self._evaluate_accuracy(train_data, net, posterior_mean)
        self.train_acc.append(np.asscalar(train_accuracy))
        if self.verbose:
            logging.info("Epoch {}. Train Loss: {}, Train_acc {}" .format(epoch, self.average_loss[-1],
                                                                          self.train_acc[-1]))

        if test_data is not None:
            test_accuracy = self._evaluate_accuracy(test_data, net, posterior_mean)
            self.test_acc.append(np.asscalar(test_accuracy))
            if self.verbose:
                logging.info("Epoch {}. Train Loss: {}, Train_acc {}, Test_acc {}".format(epoch, self.average_loss[-1],
                                                                                          self.train_acc[-1],
                                                                                          self.test_acc[-1]))

    def _train_model_from_features(self, features, labels, feature_indices_per_layer=None):
        """
        Train a BNN model using features extracted from source neural network.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :param labels: Labels to use for training.
        :type labels: :class:`numpy.ndarray`
        :param feature_indices_per_layer: Mapping of feature_layer_names to indices in features array
        i.e. {layer_name, feature_indices} Note that this param is currently not consumed by bnn_repurposer.
        :type feature_indices_per_layer: OrderedDict[str, :class:`numpy.ndarray`]
        :return: BNN model trained with given features and labels.
        :rtype: :class: `xfer.bnn_classifier.BnnClassifier`
        """

        # Load data
        x_tr = features.astype(np.dtype(np.float32))
        x_tr = self.normalizer.fit_transform(x_tr)
        y_tr = labels.astype(np.dtype(np.float32))

        num_classes = len(set(y_tr))

        train_data = gluon.data.DataLoader(gluon.data.ArrayDataset(x_tr, y_tr), batch_size=self.batch_size,
                                           shuffle=True)
        num_train = sum([self.batch_size for _ in train_data])

        # Build net and initialize parameters
        net, shapes = self._build_nn(train_data, num_classes)

        # Create Prior, variational posterior and observation model
        prior = GaussianPrior([np.array([[0.0]])], [np.array([[self.sigma]])], shapes, self._context_bnn)
        var_posterior = GaussianVariationalPosterior(mean_init=consts.MEAN_INIT_POSTERIOR,
                                                     sigma_init=consts.SIGMA_INIT_POSTERIOR, shapes=shapes,
                                                     ctx=self._context_bnn)
        obs_model = Categorical(self._context_bnn)

        # Build loss (negative of the ELBO)
        bnn_loss = BNNLoss(prior=prior, obs_model=obs_model, var_posterior=var_posterior)

        # Training

        trainer = gluon.Trainer(params=var_posterior.get_params_list() + prior.get_params_list(),
                                optimizer=consts.BNN_OPTIMIZER, optimizer_params={'learning_rate': self.learning_rate})
        smoothing_constant = .01
        moving_loss = 0.0
        for e in range(self.num_epochs):
            self._update_annealing_weight(e)

            curr_loss_total = 0.0
            for i, (data, label) in enumerate(train_data):
                data = data.as_in_context(self._context_bnn).reshape((-1, data.shape[1]))
                label = label.as_in_context(self._context_bnn)
                label_one_hot = nd.one_hot(label, num_classes)

                loss = self._compute_loss_grad(data, label_one_hot, num_train, net, bnn_loss)

                trainer.step(data.shape[0])

                # calculate moving loss for monitoring convergence
                curr_loss = nd.mean(loss).asscalar()
                curr_loss_total = curr_loss_total + curr_loss
                moving_loss = (curr_loss if ((i == 0) and (e == 0))
                               else (1 - smoothing_constant) * moving_loss + smoothing_constant * curr_loss)

            self.anneal_weights.append(self.annealing_weight)
            self.moving_loss_total.append(moving_loss)
            self.current_loss_total.append(curr_loss)
            self.average_loss.append(curr_loss_total / float(num_train))

            # Monitor accuracy on training set and test set (if provided)
            #: as self.verbose in the constructor
            self._monitor_acc(e, train_data, net, var_posterior.get_mean())

        return BnnClassifier(net, var_posterior, self.normalizer)

    def _predict_probability_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use xfer.bnn_classifier.BnnClassifier's predict method and return predicted probabilities.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted probabilities.
        :rtype: :class:`numpy.ndarray`
        """
        return self.target_model.predict(features, self.num_samples_mc_prediction, context=self._context_bnn)[1]

    def _predict_label_from_features(self, features):
        """
        Run predictions with target_model on features extracted from source neural network.
        Use xfer.bnn_classifier.BnnClassifier's predict method and return predicted labels.

        :param features: Features extracted from source neural network.
        :type features: :class:`numpy.ndarray`
        :return: Predicted labels.
        :rtype: :class:`numpy.ndarray`
        """
        return self.target_model.predict(features, self.num_samples_mc_prediction, context=self._context_bnn)[0]

    def serialize(self, file_prefix):
        """
        Saves repurposer (excluding source model) to file_prefix.json, file_prefix_posterior.json,
        file_prefix_posterior_params.npz.

        :param str file_prefix: Prefix to save file with.
        """
        output_dict = {}
        output_dict[repurposer_keys.PARAMS] = self.get_params()
        output_dict.update(self._get_attributes())

        utils.save_json(file_prefix, output_dict)

        self.target_model.var_posterior.save(file_prefix + serialization_constants.POSTERIOR_SUFFIX)

    def deserialize(self, input_dict):
        """
        Uses dictionary to set attributes of repurposer.

        :param dict input_dict: Dictionary containing values for attributes to be set to.
        """
        # Set attributes of the repurposer from input_dict
        self._set_attributes(input_dict)

        # Deserialize the target bnn model
        var_posterior = GaussianVariationalPosterior.load(input_dict[serialization_keys.FILE_PATH] +
                                                          serialization_constants.POSTERIOR_SUFFIX)
        net = self._get_net(self.num_classes)

        # Make a forward pass to initialize parameter shapes for the network
        net(mx.ndarray.zeros(shape=(self.batch_size, self.dim_input)))

        self.target_model = BnnClassifier(net, var_posterior, self.normalizer)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
"""
Model Handler constants
"""

from enum import Enum

OPERATION = 'op'
NO_OP = 'null'
LAYER_NAME = 'name'
NODES = 'nodes'
ELEMENTS_OFFSET = 'elements_offset'
PREV_SYMBOLS = 'prev_symbols'
ID2NAME = 'id2name'
LABEL_IDX = -2
DATA = 'data'
N = 'n'
SELF = 'self'
LAYER = 'layer_factory_list'
NAME = 'name'
ATTRIBUTES = 'attrs'
INPUTS = 'inputs'


class LayerType(Enum):
    FULLYCONNECTED = 'FullyConnected'
    CONVOLUTION = 'Convolution'
    ACTIVATION = 'Activation'
    POOLING = 'Pooling'
    FLATTEN = 'Flatten'
    SOFTMAXOUTPUT = 'SoftmaxOutput'
    DROPOUT = 'Dropout'
    CONCAT = 'Concat'
    BATCHNORM = 'BatchNorm'
    PLUS = '_Plus'
    SVMOUTPUT = 'SVMOutput'
    LRN = 'LRN'
    DATA = 'Data'
    MEAN = 'mean'
    EMBEDDING = 'Embedding'



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
"""
Exceptions for Model Handler
"""


class ModelError(Exception):
    """Exception type for errors caused by invalid model actions."""
    pass


class ModelArchitectureError(Exception):
    """Exception type for errors caused when model architecture is incorrect or mismatched"""
    pass


def _handle_mxnet_error(error):
    error_string = error.__str__()
    try:
        if 'Check failed: assign(&dattr, (*vec)[i]) Incompatible attr in node' in error_string:
            start_str1 = 'expected ['
            start_index1 = error_string.index(start_str1) + len(start_str1)
            end_str1 = '], got'
            end_index1 = error_string.index(end_str1)

            start_str2 = '], got ['
            start_index2 = error_string.index(start_str2) + len(start_str2)
            end_str2 = ']\n\nStack trace returned '
            end_index2 = error_string.index(end_str2)

            correct_shape = error_string[start_index1:end_index1]
            actual_shape = error_string[start_index2:end_index2]
            raise ModelArchitectureError('Weight shape mismatch: Expected shape=({}), Actual shape=({}). This can be '
                                         'caused by incorrect layer shapes or incorrect input data shapes.'.format(
                                          correct_shape, actual_shape))
    except ValueError:
        raise error
    raise error



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx
import numpy as np
from abc import abstractmethod

from .repurposer import Repurposer
from .constants import serialization_keys, repurposer_keys, serialization_constants
from .constants import neural_network_repurposer_keys as keys
from .constants import neural_network_repurposer_constants as consts
from .model_handler import ModelHandler
from . import utils


class NeuralNetworkRepurposer(Repurposer):
    """
    Base class for repurposers that create a target neural network from a source neural network through Transfer
    Learning.

    - Transfer layer architecture and weights from a source neural network (Transfer) and
    - Train a target neural network adapting the transferred network to new data set (Learn)

    :param source_model: Source neural network to do transfer leaning from
    :type source_model: :class:`mxnet.mod.Module`
    :param context_function: MXNet context function that provides device type context
    :type context_function: function
    :param int num_devices: Number of devices to use to train target neural network
    :param int batch_size: Size of data batches to be used for training the target neural network
    :param int num_epochs: Number of epochs to be used for training the target neural network
    :param str optimizer: Optimizer required by MXNet to train target neural network. Default: 'sgd'
    :param optimizer_params: Optimizer params required by MXNet to train target neural network.
           Default: {'learning_rate': 1e-3}
    :type optimizer_params: dict(str, float)
    """
    def __init__(self, source_model: mx.mod.Module, context_function=mx.context.cpu, num_devices=1, batch_size=64,
                 num_epochs=5, optimizer='sgd', optimizer_params=None):
        super(NeuralNetworkRepurposer, self).__init__(source_model, context_function, num_devices)
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self._save_source_model_default = False
        self.optimizer = optimizer
        self.optimizer_params = optimizer_params
        if self.optimizer_params is None:
            self.optimizer_params = {keys.LEARNING_RATE: consts.DEFAULT_LEARNING_RATE}

    def get_params(self):
        """
        Get parameters of repurposer that are in the constructor

        :rtype: dict
        """
        param_dict = super(NeuralNetworkRepurposer, self).get_params()
        param_dict[keys.OPTIMIZER] = self.optimizer
        param_dict[keys.OPTIMIZER_PARAMS] = self.optimizer_params
        param_dict[keys.BATCH_SIZE] = self.batch_size
        param_dict[keys.NUM_EPOCHS] = self.num_epochs
        return param_dict

    def _get_attributes(self):
        """
        Get parameters of repurposer not in constructor

        :rtype: dict
        """
        attr_dict = super()._get_attributes()
        if self.target_model is not None:
            attr_dict[serialization_keys.LAST_LAYER_NAME_TARGET] = ModelHandler(self.target_model).layer_names[-1]
        return attr_dict

    def repurpose(self, train_iterator: mx.io.DataIter):
        """
        Train a neural network by transferring layers/weights from source_model
        Set self.target_model to the repurposed neural network

        :param train_iterator: Training data iterator to use to extract features from source_model
        :type train_iterator: :class:`mxnet.io.DataIter`
        """

        # Validate the repurpose call
        self._validate_before_repurpose()

        # Prepare target model symbol (add or remove layers/ transfer weights/ freeze/ fine tune/ etc.)
        model = self._create_target_module(train_iterator)

        # Reset iterator before using it to train
        train_iterator.reset()

        # Train the model using given training data in batches
        model.fit(train_iterator,
                  optimizer=self.optimizer,
                  optimizer_params=self.optimizer_params,
                  kvstore='device',
                  eval_metric='acc',
                  allow_missing=True,
                  batch_end_callback=mx.callback.Speedometer(self.batch_size),
                  initializer=mx.init.Xavier(rnd_type='gaussian', factor_type='in', magnitude=2),
                  num_epoch=self.num_epochs)

        # Set self.target_model to the repurposed model
        self.target_model = model

        # Save data_shape and label_shapes
        self.provide_data = train_iterator.provide_data
        self.provide_label = train_iterator.provide_label

    def predict_probability(self, test_iterator: mx.io.DataIter):
        """
        Perform predictions on test data using the target_model i.e. repurposed neural network

        :param test_iterator: Test data iterator to return predictions for
        :type test_iterator: :class:`mxnet.io.DataIter`
        :return: Predicted probabilities
        :rtype: :class:`numpy.ndarray`
        """

        # Validate the predict call
        self._validate_before_predict()

        # Bind target model symbol with data if not already bound
        if not self.target_model.binded:
            self.target_model.bind(data_shapes=test_iterator.provide_data, for_training=False)

        # Call mxnet.BaseModule.predict. It handles batch padding and resets iterator before prediction.
        predictions = self.target_model.predict(eval_data=test_iterator)
        return predictions.asnumpy()

    def predict_label(self, test_iterator: mx.io.DataIter):
        """
        Perform predictions on test data using the target_model i.e. repurposed neural network

        :param test_iterator: Test data iterator to return predictions for
        :type test_iterator: mxnet.io.DataIter
        :return: Predicted labels
        :rtype: :class:`numpy.ndarray`
        """

        # Call predict_probability method to get soft predictions
        probabilities = self.predict_probability(test_iterator)

        # Select and return label with maximum probability for each test instance
        labels = np.argmax(probabilities, axis=1)

        return labels

    @abstractmethod
    def _create_target_module(self, train_iterator: mx.io.DataIter):
        """
        Abstract method.
        """
        pass

    def _validate_before_predict(self):
        # For neural network repurposers, target_model should be an MXNet module object
        if not isinstance(self.target_model, mx.mod.Module):
            error = ("Cannot predict because target_model is not an `mxnet.mod.Module` object. "
                     " Instead got type: {}".format(type(self.target_model)))
            raise TypeError(error)

        if not self.target_model.params_initialized:
            error = "target_model params aren't initialized. Ensure model is trained before calling predict"
            raise ValueError(error)

    def serialize(self, file_prefix):
        """
        Serialize repurposer to dictionary

        :return: Dictionary describing repurposer
        :rtype: dict
        """
        output_dict = {}
        output_dict[repurposer_keys.PARAMS] = self.get_params()
        output_dict.update(self._get_attributes())

        # save serialised model file_path.json
        utils.save_json(file_prefix, output_dict)

        # Save target_model as file-path-symbol.json and file-path-0000.params
        utils.save_mxnet_model(self.target_model, file_prefix, 0, self.provide_data, self.provide_label)

    def deserialize(self, input_dict):
        """
        Uses dictionary to set attributes of repurposer

        :param dict input_dict: Dictionary containing values for attributes to be set to
        """
        self.target_model = mx.mod.Module.load(input_dict[serialization_keys.FILE_PATH], 0,
                                               label_names=[input_dict[serialization_keys.LAST_LAYER_NAME_TARGET] +
                                                            serialization_constants.LABEL_SUFFIX])
        self._set_attributes(input_dict)  # Set attributes of repurposer from input_dict



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import numpy as np
import os
import mxnet as mx
import json

from .constants import serialization_constants as consts
from .constants import repurposer_keys as keys


def sklearn_model_to_dict(target_model):
    output_dict = {}
    import copy
    # model_dict contains all attributes of model
    model_dict = copy.deepcopy(target_model.__dict__)
    for k in model_dict:
        # Replace any numpy array with [data_type_as_str, array_as_list]
        # e.g np.array([1,2]) -> ['int', [1,2]]
        if isinstance(model_dict[k], np.ndarray):
            type_data = str(model_dict[k].dtype)
            model_dict[k] = [type_data, model_dict[k].tolist()]
        # Replace any tuple with ['tuple', tuple_as_list]
        # e.g (1,2) -> ['tuple', [1,2]]
        if isinstance(model_dict[k], tuple):
            model_dict[k] = [keys.TUPLE, list(model_dict[k])]
    output_dict[keys.MODEL] = {}
    # Model params are public attributes
    output_dict[keys.MODEL][keys.PARAMS] = target_model.get_params()
    # Serialise all private attributes
    output_dict[keys.MODEL][keys.ATTRS] = {}
    for k in model_dict:
        # Serialize private parameters as attributes
        if k[-1] == '_' or k[0] == '_':
            output_dict[keys.MODEL][keys.ATTRS][k] = model_dict[k]
    return output_dict


def sklearn_model_from_dict(model_class, input_dict):
    # Initialize model with serialized model parameters
    model = model_class(**input_dict[keys.MODEL][keys.PARAMS])
    # Set model attributes
    for k in input_dict[keys.MODEL][keys.ATTRS]:
        # Unpack tuples and np.arrays that were serialised as lists
        if isinstance(input_dict[keys.MODEL][keys.ATTRS][k], list) \
           and isinstance(input_dict[keys.MODEL][keys.ATTRS][k][0], str) \
           and type(input_dict[keys.MODEL][keys.ATTRS][k][1]) == list:
            if input_dict[keys.MODEL][keys.ATTRS][k][0] == keys.TUPLE:
                setattr(model, k, tuple(input_dict[keys.MODEL][keys.ATTRS][k][1]))
            else:
                type_data = 'np.' + input_dict[keys.MODEL][keys.ATTRS][k][0]
                type_data = eval(type_data)
                setattr(model, k, np.array(input_dict[keys.MODEL][keys.ATTRS][k][1], dtype=type_data))
        else:
            setattr(model, k, input_dict[keys.MODEL][keys.ATTRS][k])
    return model


def _assert_repurposer_file_exists(repurposer_file_list):
    for file_name in repurposer_file_list:
        if not os.path.isfile(file_name):
            raise NameError('Cannot find repurposer file ({})'.format(file_name))


def save_mxnet_model(model, file_path_prefix, epoch, provide_data=None, provide_label=None):
    if not model.binded:
        if provide_data is None or provide_label is None:
            raise ValueError("provide_data and provide_label are required because mxnet module is not binded")
        model.bind(data_shapes=provide_data, label_shapes=provide_label)
    model.save_checkpoint(file_path_prefix, epoch)


def save_json(file_prefix, output_dict):
    with open(file_prefix + consts.JSON_SUFFIX, mode='w') as fp:
        json.dump(obj=output_dict, fp=fp)


def serialize_ctx_fn(context_function):
    if context_function == mx.cpu:
        return keys.CPU
    elif context_function == mx.gpu:
        return keys.GPU
    else:
        raise ValueError('Unexpected context function {}'.format(context_function))


def deserialize_ctx_fn(context_function):
    if context_function == keys.CPU:
        return mx.cpu
    elif context_function == keys.GPU:
        return mx.gpu
    else:
        raise ValueError('Unexpected context function {}'.format(context_function))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
from setuptools import setup, find_packages
import sys


with open('README.md', 'r') as fh:
    long_description = fh.read()


requires = [
    'mxnet==1.2.0',
    'numpy==1.14.5',
    'scikit-learn==0.19.2',
    'matplotlib==2.2.2',
    'GPy==1.9.5'
]


# get __version__ variable
exec(open("xfer/__version__.py").read())


# enforce >Python3 for all versions of pip/setuptools
assert sys.version_info >= (3,), 'This package requires Python 3.'


setup(
    name='xfer-ml',
    version=__version__,  # noqa
    description='Lightweight deep transfer learning library built on MXNet',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/amzn/xfer',
    packages=find_packages(exclude=['test*']),
    include_package_data=True,
    install_requires=requires,
    python_requires='>=3',  # requires setuptools>=24.2.0 for packaging and pip>=9.0.0 on download for this to work
    license='Apache License 2.0',
    classifiers=(
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    )
)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import os
import json
import mxnet as mx

from . import utils
from .constants import repurposer_keys, bnn_repurposer_keys
from .constants import serialization_constants as consts
from .constants import serialization_keys as keys

# Import every repurposer
from .lr_repurposer import LrRepurposer  # noqa
from .svm_repurposer import SvmRepurposer  # noqa
from .bnn_repurposer import BnnRepurposer
from .gp_repurposer import GpRepurposer  # noqa
from .neural_network_fine_tune_repurposer import NeuralNetworkFineTuneRepurposer  # noqa
from .neural_network_random_freeze_repurposer import NeuralNetworkRandomFreezeRepurposer  # noqa


def load(model_name, model_directory='', source_model=None):
    """
    Load the repurposed model (target_model and supporting info) from given file_path and deserialize

    :param str model_name: Name of saved repurposer
    :param str model_directory: Directory of saved repurposer
    :param source_model: Source neural network to do transfer learning from. If this is None, then source model
                         (model_name_source-symbol.json & model_name_source-0000.params) will be loaded from file
    :type source_model: :class:`mxnet.mod.Module`
    :return: Repurposer object loaded from file
    :rtype: :class:`Repurposer` (Object returned will be resolved to originally saved class)
    """
    file_prefix = os.path.join(model_directory, model_name)
    utils._assert_repurposer_file_exists([file_prefix + consts.JSON_SUFFIX])

    with open(file_prefix + consts.JSON_SUFFIX, mode='r') as json_data:
        input_dict = json.load(json_data)

    # load source model from file if required
    if source_model is None:
        utils._assert_repurposer_file_exists([file_prefix + '_source-symbol.json', file_prefix + '_source-0000.params'])
        source_model = mx.module.Module.load(file_prefix + consts.SOURCE_SUFFIX, 0,
                                             label_names=[input_dict[keys.LAST_LAYER_NAME_SOURCE] +
                                                          consts.LABEL_SUFFIX])

    # Add file_prefix to input dict
    input_dict[keys.FILE_PATH] = file_prefix

    repurposer_class = eval(input_dict[repurposer_keys.REPURPOSER_CLASS])

    # Deserialize input dict
    context_function = utils.deserialize_ctx_fn(input_dict[repurposer_keys.PARAMS][repurposer_keys.CONTEXT_FN])
    input_dict[repurposer_keys.PARAMS][repurposer_keys.CONTEXT_FN] = context_function

    if repurposer_class == BnnRepurposer:
        bnn_context_function = utils.deserialize_ctx_fn(input_dict[repurposer_keys.PARAMS]
                                                                  [bnn_repurposer_keys.BNN_CONTEXT_FUNCTION])
        input_dict[repurposer_keys.PARAMS][bnn_repurposer_keys.BNN_CONTEXT_FUNCTION] = bnn_context_function

    # Instantiate repurposer and deserialize input dictionary
    repurposer = repurposer_class(source_model, **input_dict[repurposer_keys.PARAMS])
    repurposer.deserialize(input_dict)
    return repurposer



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import mxnet as mx
from mxnet import gluon
import numpy as np

from .utils import serialise_ctx, WRITE, NULL, SHAPE, CONTEXT


class Probability(object):

    """
    Base class for defining a probability over a set of variables [W_1, ..., W_n] where W_i is an mxnet.ndarray.

    :param params: Dictionary containing the parameters that characterize the probability distribution
      (e.g. mean and variance of each W_i when the Probability represents a gaussian distribution)
    :type params: dict(str, :class:`mxnet.gluon.Parameter`)
    :param raw_params: Dictionary containing the values of the parameters stored in params
    :type raw_params: dict(str, list[:class:`mxnet.ndarray`]}
    :param shapes: Shape of the variables over which the probability is defined.
    :type shapes: list(tuple(int, int))
    :param ctx: MXNet context
    :type ctx: :class:`mxnet.Context`
    """

    def __init__(self, shapes, ctx):

        """
        Default constructor

        :param shapes: Shape of the variables over which the probability is defined.
        :type shapes: list(tuple(int, int))
        :param ctx: MXNet context
        :type ctx: :class:`mxnet.Context`
        """

        self.ctx = ctx
        self.params = {}
        self.raw_params = {}
        self.shapes = shapes

    def get_params_list(self):
        # Each element of the dictionary contains a parameter (list(nd.array)). This method
        # iterates over all the elements in the dictionary and outputs a single list(nd.array) containing
        # all parameters. The dictionary is sorted beforehand so the output is deterministic and the
        # method can be tested.
        return [x
                for kk in sorted(self.params)
                for x in self.params[kk]
                ]

    def check_observation_shapes(self, obs):
        shape_obs = [o.shape for o in obs]
        if not self.shapes == shape_obs:
            raise ValueError("Number of observed variables {} does not match \
             the number of expected observed variables {}".format(shape_obs, self.shapes))

    def log_pdf(self, x):
        raise NotImplemented

    def KL(self, other_prob):
        """
        Function to compute the KL divergence between self and a second Probability

        :param other_prob: The probability object that we want to compute th KL distance to
        :type other_prob: :class:`xfer.prob.Probability`
        """
        raise NotImplemented

    def KL_sample(self, other_prob, sample):
        """
        Function to compute the KL divergence between a sample of self and a sample of a second Probability

        :param other_prob: The probability that we want to compute the KL distance to
        :type other_prob: :class:`xfer.prob.Probability`
        :param sample: Sample used to compute the KL
        list[:class:`np.ndarray`]
        """
        return self.log_pdf(sample) - other_prob.log_pdf(sample)

    def _parse_grad_req(self, fix, num_parameters=None):
        """
        :param fix: define for each parameter whether the gradient is computed  (True) or
            not (False). If it is a scalar, the value is replicated num_parameters times.
        :type fix: boolean or list(boolean)
        :return: define for each parameter whether the gradient is computed  ("write") or not ("null")
        :rtype: list[string]
        """
        if isinstance(fix, list):
            grad_req = [NULL if f else WRITE for f in fix]
        elif fix:
            grad_req = [NULL] * num_parameters
        else:
            grad_req = [WRITE] * num_parameters
        return grad_req

    def _register_param(self, name, init, fix=False):
        """
        Register a set of parameters using "init".

        Its shape is equal to the shape of the variables over which the probability is defined (self.shapes).
        In addition to updating the parameters' dictionary (self.params), this function updates a dictionary with the
        value of the parameters (self.raw_params).

        :param string name: Name of the set of parameters.
        :param init: Define how to initialize the set of parameters
        :type init: mxnet.Initializer
        :param fix: define for each parameter whether the gradient is computed  (True) or not (False). If it is a single
         boolean it is applied globally to the full set of variables.
        :type fix: (boolean or list(boolean))
        """

        par_list = []
        grad_req = self._parse_grad_req(fix, len(self.shapes))
        for shape, gg in zip(self.shapes, grad_req):
            par = gluon.Parameter(name, shape=shape, init=init, grad_req=gg)
            par.initialize(ctx=self.ctx)
            par_list.append(par)
        self.params[name] = par_list
        self.raw_params[name] = [x.data(self.ctx) for x in self.params[name]]

    def _register_param_value(self, name, values, fix=False):
        """
        Register a parameter and initialize it with values (list(nd.ndarray)).

        It also checks whether the shape (given by "values") is compatible with the shape of the variables
        over which the probability is defined (self.shapes).

        :param string name: Name of the set of parameters.
        :param values: Value of the set of parameters.
                [np.array]: Parameters shared across all variables.
                [np.array_1, ..., np.array_n]: Different parameters for each variable W_i
        :type values: list[:class:`np.ndarray`]
        :param fix: define for each parameter whether the gradient is computed  (True) or not (False). If it is a single
         boolean it is applied globally to the full set of variables.
        :type fix: (boolean or list(boolean))
        """
        values_shapes = [vv.shape for vv in values]

        if not len(values) == 1:
            if not (len(values_shapes) == len(self.shapes)):
                raise ValueError("Parameter {} size is not compatible".format(name))
            for value_shape, parameter_shape in zip(values_shapes, self.shapes):
                if not (value_shape == parameter_shape or value_shape == (1, 1)):
                    raise ValueError("Parameter {} size is not compatible".format(name))
        else:
            if not values[0].shape == (1, 1):
                raise ValueError("Parameter {} size is not compatible".format(name))

        par_list = []
        grad_req = self._parse_grad_req(fix, len(values))
        for vv, gg in zip(values, grad_req):
            init = mx.init.Constant(vv)
            par = gluon.Parameter(name, shape=vv.shape, init=init, grad_req=gg)
            par.initialize(ctx=self.ctx)
            par_list.append(par)
        self.params[name] = par_list
        self.raw_params[name] = [x.data(self.ctx) for x in self.params[name]]

    def _replicate_shared_parameters(self):
        """
        It returns a dictionary containing the values of each set of parameters of the distribution.

        The difference with self.raw_params is that the shared parameters are replicated.
        """
        raw_params_ext = {}
        for k in self.raw_params:
            if len(self.raw_params[k]) == 1:
                # If the parameter is shared, replicate it len(self.shapes) times
                raw_params_ext[k] = [self.raw_params[k][0] for _ in range(len(self.shapes))]
            else:
                raw_params_ext[k] = self.raw_params[k]
        return raw_params_ext

    def _to_dict(self):
        output_dict = {
            SHAPE: self.shapes,
            CONTEXT: serialise_ctx(self.ctx)
        }
        return output_dict

    def _save_params(self, filename):
        np.savez(filename, **self.params)

    @staticmethod
    def _load_params(filename):
        load_data = np.load(filename)
        params = {}
        raw_params = {}
        for key in load_data.files:
            params[key] = load_data[key]
            raw_params[key] = []
            for parameter in params[key]:
                raw_params[key].append(parameter.data())

        return params, raw_params



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
import numpy as np
import mxnet as mx
import json
import mxnet.ndarray as nd

from .prior import Probability, GaussianPrior
from .utils import sample_epsilons, transform_rhos, log_gaussian, transform_gaussian_samples, softplus_inv_numpy,\
    deserialise_ctx, deserialise_shape, MEAN, RHO, PARAMS, JSON, CONTEXT, SHAPE, NPZ


class VariationalPosterior(Probability):
    def __init__(self, shapes, ctx):
        super(VariationalPosterior, self).__init__(shapes, ctx)

    def generate_sample(self):
        pass

    def get_mean(self):
        pass


class GaussianVariationalPosterior(VariationalPosterior):
    """
    Gaussian variational distribution (iid) over a set of variables [W_1, ..., W_n] where W_i is an mxnet.array

    :param params: Dictionary containing the parameters that characterize the probability distribution
      , i.e. mean and variance of each W_i.
    :type params: dict(str, :class:`mxnet.gluon.Parameter`)
    :param raw_params: Dictionary containing the values of the parameters stored in params
    :type raw_params: dict(str, list[:class:`mxnet.ndarray`]}
    :param shapes: Shape of the variables over which the probability is defined.
    :type shapes: list(tuple(int, int))
    :param ctx: MXNet context
    :type ctx: :class:`mxnet.Context`
    """

    def __init__(self, mean_init, sigma_init, shapes, ctx, fix_means=False, fix_sigmas=False):
        """
        Default constructor

        :param mean_init: The initial variational mean.
            float: Different mean for each value of each W_i (sampled from N(mean_init, 1))
            [np.array]: Mean shared across all variables.
            [np.array_1, ..., np.array_n]: Different mean for each variable W_i
        :type mean_init: float or list[:class:`np.ndarray`]
        :param sigma_init: Initial variational variance.
            float: Different variance for each variable (initialize to float value)
            [np.array]: Variance shared across all variables.
            [np.array_1, ..., np.array_n]: Different variance for each variable W_i
        :type sigma_init: float or list[:class:`np.ndarray`]
        :param shapes: Size of the variables [W_1, ..., W_n].
        :type shapes: list(tuple(int, int))
        :param ctx: MXNet context
        :type ctx: :class:`mxnet.Context`
        :param fix_means: Fix the mean of the Variational Posterior.
        :type fix_means: boolean or list(boolean)
        :param fix_sigmas: Fix the variance of the Variational Posterior.
        :type fix_sigmas: boolean or list(boolean)
        """

        super(GaussianVariationalPosterior, self).__init__(shapes, ctx)

        self._register_params(mean_init, sigma_init, fix_means, fix_sigmas)

    def _register_params(self, mean_init, sigma_init, fix_means, fix_sigmas):
            if not isinstance(mean_init, list):
                self._register_param(MEAN, mx.init.Normal(mean_init), fix_means)
            else:
                self._register_param_value(MEAN, mean_init, fix_means)
            if not isinstance(sigma_init, list):
                rho_init = softplus_inv_numpy(sigma_init)
                self._register_param(RHO, mx.init.Constant(rho_init), fix_sigmas)
            else:
                rho_init = [softplus_inv_numpy(ss) for ss in sigma_init]
                self._register_param_value(RHO, rho_init, fix_sigmas)

    def get_mean(self):
        return self.raw_params[MEAN]

    def generate_sample(self):
        epsilons = sample_epsilons(self.shapes, self.ctx)
        raw_params_ext_var_posterior = self._replicate_shared_parameters()
        sigmas = transform_rhos(raw_params_ext_var_posterior[RHO])
        return transform_gaussian_samples(raw_params_ext_var_posterior[MEAN], sigmas, epsilons)

    def log_pdf(self, obs):
        self.check_observation_shapes(obs)
        raw_params_ext = self._replicate_shared_parameters()
        sigmas = transform_rhos(raw_params_ext[RHO])
        return sum([nd.sum(log_gaussian(obs[ii], raw_params_ext[MEAN][ii], sigmas[ii]))
                    for ii in range(len(self.shapes))])

    def is_conjugate(self, other_prob):
        if type(other_prob) == GaussianPrior:
            return True
        else:
            return False

    def KL(self, other_prob):
        if not self.is_conjugate(other_prob):
            raise ValueError("KL cannot be computed in closed form.")

        if (not len(self.shapes) == len(other_prob.shapes)) or \
                (not np.all(np.array([s == o for s, o in zip(self.shapes, other_prob.shapes)]))):
            raise ValueError("KL cannot be computed: The 2 distributions have different support")

        raw_params_ext_var_posterior = self._replicate_shared_parameters()
        sigmas_var_posterior = transform_rhos(raw_params_ext_var_posterior[RHO])
        raw_params_ext_prior = other_prob._replicate_shared_parameters()

        out = 0.0
        for ii in range(len(self.shapes)):
            means_p = raw_params_ext_prior[MEAN][ii]
            var_p = raw_params_ext_prior["sigma"][ii] ** 2
            means_q = raw_params_ext_var_posterior[MEAN][ii]
            var_q = sigmas_var_posterior[ii] ** 2
            inc_means = (means_q - means_p)
            prec_p = 1.0 / var_p
            temp = 0.5 * (var_q*prec_p + ((inc_means ** 2) * prec_p) - 1.0 + nd.log(var_p) - nd.log(var_q))
            if temp.shape == (1, 1):
                # If parameters are shared, multiply by the number of variables
                temp = temp * (self.shapes[ii][0] * self.shapes[ii][1])
            out = out + nd.sum(temp)
        return out

    def save(self, filename):
        """
        Save object to file. This will create two files: filename.json, filename_params.npz

        :param filename: Name to give saved object files
        """
        self._save_params(filename + PARAMS)
        output_dict = self._to_dict()
        with open(filename + JSON, 'w') as fp:
            json.dump(obj=output_dict, fp=fp)

    @staticmethod
    def load(filename):
        """
        Load object from file. This requires two files: filename.json, filename_params.npz

        :param filename: Names of files to be opened and loaded
        """
        with open(filename + JSON, 'r') as json_data:
            input_dict = json.load(json_data)
        ctx = deserialise_ctx(input_dict[CONTEXT])
        shape = deserialise_shape(input_dict[SHAPE])

        gvp = GaussianVariationalPosterior(1, 1, shape, ctx, False, False)

        params, raw_params = GaussianVariationalPosterior._load_params(filename + PARAMS + NPZ)
        gvp.params = params
        gvp.raw_params = raw_params

        return gvp
